
/* Each 8086 register is 2 bytes */

#ifndef EXTERN
#define EXTERN extern
#endif 

#ifndef INIT
#define INIT(x)
#endif

#include "regtypes.h"
#include "misc.h"


/* need to do something here about endian -- */
typedef union {
	WORD_REG wordreg;
	struct  {
		BYTE_REG high;
		BYTE_REG low;
	} bytereg; 
} REG;	

typedef struct {
	REG ax_reg;
	REG bx_reg;
	REG cx_reg;
	REG dx_reg;
	WORD_REG si; 
	WORD_REG di;
	WORD_REG sp;
	WORD_REG bp;
	WORD_REG cs;
	WORD_REG ip;
	WORD_REG ds;
	WORD_REG es;	
	WORD_REG ss;
} REGS_8086;
 
	
#define al	ax_reg.bytereg.low
#define ah	ax_reg.bytereg.high
#define bl	bx_reg.bytereg.low
#define bh  	bx_reg.bytereg.high
#define cl 	cx_reg.bytereg.low
#define ch	cx_reg.bytereg.high
#define dh	dx_reg.bytereg.high
#define dl	dx_reg.bytereg.low
#define ax	ax_reg.wordreg
#define bx	bx_reg.wordreg
#define cx	cx_reg.wordreg
#define dx	dx_reg.wordreg

/* if w== 1 */
#define AX	0
#define BX	3
#define CX	1
#define DX	2
#define SP	4
#define BP	5
#define SI	6
#define DI 	7

/* if w== 0 */
#define AL	0
#define CL	1
#define DL	2
#define BL 	3
#define AH	4
#define CH	5
#define DH	6
#define BH	7



/* segment registers */
#define ES 	0
#define CS	(1 << 3)
#define SS	(2 << 3)
#define DS	(3 << 3)


typedef struct {
	BOOLEAN carry;	/* bit 0 */
	BOOLEAN parity;	/*  bit 2 */
	BOOLEAN aux_carry;	/* auxillary carry, bit 4 */
	BOOLEAN zero;	/* zero flag, bit 6 */
	BOOLEAN sign;	/* sign flag, bit 7 */
	BOOLEAN overflow;	/* overflow flag, bit 11 */
	BOOLEAN trap;	/* bit 8 */
	BOOLEAN interrupt_enable; 	/* bit 9 */
	BOOLEAN direction;	/* bit 10 */
} FLAGS;


/* segment selectors actually point into pc ram */
EXTERN BYTE *fast_cs;
EXTERN BYTE *fast_ss;
EXTERN BYTE *fast_es;
EXTERN BYTE *fast_ds;

EXTERN BYTE pc_ram[1024 * 1024];        /* emulate 640 k of PC ram */
EXTERN  REGS_8086 regs;
EXTERN REGS_8086 boot_regs;
EXTERN  FLAGS flags;

EXTERN int num_instructions INIT(0);
EXTERN char disassemble_string[40];	/* null terminated array for dissembled instructions */

typedef enum {  PC_INTERRUPT, EMULATED_INTERRUPT } TYPE_OF_INTR;

/* If no interrupt is assigned, a PC will crash on interrupt.
 * A PC interrupt emulated a PC
 * An emulated interrupt goes to one of our functions, which acts as an
 * interrupt handler, and returns like an iret occurred [correct registers
 * and flags for the interrupt
 *
 * On an emulated interrupt, the next opcode is the next byte of the program.
 * On a PC interrupt, it act like an interrupt (next byte is first byte of the
 * interrupt handler, flags and pc far are saved on stack
 */
typedef struct {
        TYPE_OF_INTR type_int;
        void (*int_ptr)();      /* don't care for  PC_INTERRUPT,
                         * points to a host function for EMULATED_INTERRUPT
                         */
} INTERRUPT;

EXTERN INTERRUPT interrupt_table[256];

/* put this functions here, and inline them (better than a macro...
 * this is so we can use the debugger on them.
 * They're static so they won't be external and we won't have a problem
 * compiling files multiple different ways.
 */
static inline WORD read_pc_word(BYTE *address)
{
	WORD tmp;

#ifdef NO_ACCESS_INTERRUPTS
        if(address < &pc_ram[0x255*4]) {
                fprintf(stderr, "can't read interrupt tables\n");
                abort();
        }
#endif
#ifdef READ_PC_WORD_DEBUG
	tmp = *address;
	tmp += *(address + 1) << 8;
	return tmp;
#else	
        return ((WORD) *address + (((WORD) *(address + 1)) << 8));
#endif
}

static inline void write_pc_word(BYTE *address, WORD word)
{

#ifdef NO_ACCESS_INTERRUPTS
        if(address < &pc_ram[0x255*4]) {
                fprintf(stderr, "can't write interrupt tables\n");
                abort();
        }
#endif

        *address = word & 0xff;
        *(address +1) = word >> 8;
}



typedef enum {
        NO_SEG, DS_SEG, CS_SEG, ES_SEG, SS_SEG
} SEG;

static void set_selector(const SEG seg, const WORD value)
{
        void *p = (void *) &pc_ram[(int) value << 4];

        switch (seg) {
            case DS_SEG:
                    fast_ds = p;
                    regs.ds = value;
                    break;
            case ES_SEG:
                    fast_es = p;
                    regs.es = value;
                    break;
            case SS_SEG:
                    fast_ss = p ;
                    regs.ss = value;
                    break;
            case CS_SEG:
                    fast_cs = p;
                    regs.cs = value;
                    break;
	   default:	
			fprintf(stderr, "bad seg in set_selector()\n");
			break;	
        }
}


/* common sequence for multi-byte opcodes --
 * assumes there are variables:
 *      rm, mod, secondary opcode
 *      secondary byte
 */
#define get_secondary_byte() \
        { secondary_byte = fetch(); secondary_opcode = secondary_byte & SECONDARY_MASK; \
          rm = secondary_byte & 0x7; mod = (secondary_byte & 0xc0) >> 6; }
 
#define compute_effective_address_byte(mod, rm) compute_effective_address(mod, rm, TRUE)
#define compute_effective_address_word(mod, rm) compute_effective_address(mod, rm, FALSE)
 

static void *get_pcram_seg(WORD segment)
{
	return &pc_ram[segment << 4];
}


/* bit assignments -- used to emulate the hardware and internally */
#define CF  (1)
#define PF (3)
#define AF	(5)
#define ZF	(7)
#define SF	(8)
#define OF	(12)
#define TF	(9)
#define IF	(10)
#define DF	(11)	
