#include <stdio.h>

#include "8086regs.h"
#include "misc.h"

typedef struct _breakpoint {
	struct _breakpoint *next;
	struct _breakpoint *prev;
	/* unix addresses -- for fetch, watching byte or word */
	void   *address;
	enum { WATCH_WORD, WATCH_BYTE, FETCH, IN_INSTRUCTIONS, 
			WATCH_WORD_REG,
			WATCH_WORD_REG_FOR_VALUE,
			EVERY_N_INSTRUCTIONS } op;
	WORD past_word;	/* watching word */
	BYTE past_byte; /* watching byte */
	int instructions; /* counting instrucitons */
	/* keep this around for addresses */
	int n_instructions;
	WORD segment; 
	WORD offset;	
	/* whether breakpoint is active or not */
	BOOLEAN active;
} BREAKPOINT;


static inline void *make_pc_pointer(WORD segment, WORD offset)
{
        return (void *) &pc_ram[(segment << 4) + offset];
}


static BREAKPOINT *bp_list;

static BREAKPOINT *add_breakpoint(void)
{
	BREAKPOINT *bp;

	bp = (BREAKPOINT *) malloc(sizeof(BREAKPOINT));
	assert(bp);
	bp->next = bp_list;
	bp->prev = NULL;
	bp_list = bp;
	bp->active = TRUE;
	return bp;
}

static void subtract_breakpoint(BREAKPOINT *bp)
{
	if(bp_list == bp) {
		bp_list = bp->next;
		bp_list->prev = NULL;
	} else {
		(bp->next)->prev = bp->prev;
		(bp->prev)->next = bp->next;
	}
	free(bp);
	
}		

static void have_match(BREAKPOINT *bp)
{
	show_a_breakpoint(bp);	
}

static void make_word(WORD segment, WORD offset)
{
	BREAKPOINT *bp = add_breakpoint();

	bp->op = WATCH_WORD;
	bp->address = make_pc_pointer(segment, offset);
	bp->past_word = read_pc_word(bp->address);
	bp->segment = segment;
	bp->offset = offset;

}

static void make_word_reg(WORD *ptr)
{
	BREAKPOINT *bp = add_breakpoint();

	bp->op = WATCH_WORD_REG;
	bp->address = ptr;
	bp->past_word = *ptr;
}

static void make_word_reg_value(WORD *ptr, WORD value)
{
	BREAKPOINT *bp = add_breakpoint();

	bp->op =  WATCH_WORD_REG_FOR_VALUE;
	bp->address = ptr;
	bp->past_word = value;
}
		

static void make_byte(WORD segment, WORD offset)
{
	BREAKPOINT *bp = add_breakpoint();

	bp->op = WATCH_BYTE;
	bp->address = make_pc_pointer(segment, offset);
	bp->past_byte = *(BYTE *) (bp->address);
	bp->segment = segment;
	bp->offset = offset;
}

static void make_fetch(WORD segment, WORD offset)
{
	BREAKPOINT *bp = add_breakpoint();

	bp->op = FETCH;
	bp->address = make_pc_pointer(segment, offset);
	bp->segment = segment;
	bp->offset = offset;
}

static void make_instructions(int num)
{
	BREAKPOINT *bp = add_breakpoint();
	
	bp->op = IN_INSTRUCTIONS;
	bp->instructions = num_instructions + num;	
	
}

static void make_every_n_instructions(int num)
{
	BREAKPOINT *bp = add_breakpoint();

	bp->op = EVERY_N_INSTRUCTIONS;
	bp->instructions = num_instructions + num;
}

static void show_address_bp(const char *string, const BREAKPOINT *bp)
{
	printf("%s at %04x:%04x ", string, bp->segment, bp->offset);
	if(bp->op == WATCH_WORD) {
		printf("%dD 0x%x\n", read_pc_word(bp->address), read_pc_word(bp->address));
	} else if(bp->op == WATCH_BYTE) 
		printf("%dD 0x%x\n", *(BYTE *) bp->address, *(BYTE *) bp->address);
	 else  printf("\n");
}

static void disable_breakpoint(int num)
{
}

static void delete_breakpoint(int num)
{
}

static void show_a_breakpoint(BREAKPOINT *bp)
{
	printf("At %04x:%04x\t", regs.cs, regs.ip);
	switch(bp->op) {
		case FETCH:	
			show_address_bp("FETCH", bp);
			break;
		case WATCH_BYTE: 
			show_address_bp("BYTE", bp);
			break;
		case WATCH_WORD: 
			show_address_bp("WORD", bp);
			break;
		case IN_INSTRUCTIONS:
			printf("INSTRUCTION #%d, now %d\n",
				bp->instructions, num_instructions);
			break;	
		case WATCH_WORD_REG:
			printf("WORD REG at %x, now %x\n", bp->address,
				*(WORD *) bp->address);
			break;
		case WATCH_WORD_REG_FOR_VALUE:
			printf("WATCH WORD REG at %x for %x\n",
					bp->address, bp->past_word);
			break;
		default:
			assert(0);
	}
}

static void show_all_breakpoints(void)
{
	BREAKPOINT *bp;
	int i;

	for(i = 0, bp = bp_list; bp; i++, bp = bp->next) {
		printf("%d -- %x  ", i, bp);
		show_a_breakpoint(bp);
	}
}
	

static void make_breakpoint()
{
}

void test_breakpoint(void)
{
	BREAKPOINT *bp;

	num_instructions++;

#ifdef TRACE
        print_like_aztec();
        debug_trace("emulate instruction at %x:%x\n", regs.cs, regs.ip);                
        last_ip = regs.ip;
#endif

	for(bp = bp_list; bp; bp = bp->next) {
		if(bp->active == FALSE) 
			continue;
		switch(bp->op) {
			case WATCH_WORD:
				if(bp->past_word != read_pc_word(bp->address)) { 
					have_match(bp);
					bp->past_word = read_pc_word(bp->address);
				}
				break;
			case WATCH_BYTE:
				if(bp->past_byte != *(BYTE *) bp->address) { 
					have_match(bp);
					bp->past_byte = *(BYTE *) bp->address;
				}
				break;
			case FETCH:
				if((fast_cs + regs.ip ) == bp->address) {
					have_match(bp);
				}
				break;
			case IN_INSTRUCTIONS:
				if(bp->instructions == num_instructions)
					have_match(bp);
				break;
			case WATCH_WORD_REG_FOR_VALUE:
				if(bp->past_word == *(WORD *) bp->address)
					have_match(bp);
				break;
			case WATCH_WORD_REG:
				if(bp->past_word != *(WORD *) bp->address) {
					have_match(bp);
					bp->past_word = *(WORD *)bp->address;
				}
				break; 
			default:
				assert(0);
		}			

	}

} 
