/* An 8086 disassembler
 *
 */

#include <stdio.h>

#include "misc.h"
#include "8086regs.h"
#include "8086opcodes1.h"
#include "8086opcodes2.h"

/* where instruction is */
static BYTE *pc;	

/* size of instruction */
static int instruction_size;
static WORD current_ip;	/* ip for instruction */

static BYTE secondary_opcode, rm, mod;
static BYTE secondary_byte;

static SEG seg_override = NO_SEG;

#define fetch()		(*pc) 
/* number of bytes found elsewhere for opcode */
static int num_bytes;

static void process_secondary_byte(void)
{
	get_secondary_byte();
	num_bytes++;
	pc++;
}

static WORD read_next_pc_word(void)
{
	WORD temp;

	num_bytes += 2;
	temp = read_pc_word(pc);
	pc += 2;
	return temp;
}


static char *show_imm_byte(void)
{
	static char buffer[20];

	sprintf(buffer, "%x", *pc);
	num_bytes++;
	pc++; 
	return buffer;
}

static char *show_imm_word(void)
{
	static char buffer[20];

	sprintf(buffer, "%x", read_next_pc_word());
	return buffer;
}

	
static char *show_seg_override(void)
{
	switch(seg_override) {
		case CS_SEG: 	return "CS:";
		case DS_SEG:	return "DS:";
		case ES_SEG:	return "ES:";
		case SS_SEG:	return "SS:";
		case NO_SEG:	return "";	/* none */	
		default:	return "seg override wrong";
	}
}

static void set_string(const char *string)
{
	strcat(disassemble_string, string);	
}


static char *show_offset()
{
	WORD offset;
	static char buffer[20];

	switch(mod) {
		case 0:  offset = 0; break;
		case 1:	 offset = *pc; pc++; num_bytes++; break;
		case 2:  offset = read_next_pc_word(); break;
		default:	fprintf(stderr, "somethning is wrong\n");
	}

	switch(rm) {
		case 0:	sprintf(buffer, "[BX][SI]%x", offset); break;
		case 1: sprintf(buffer, "[BX][DI]%x", offset); break;
		case 2: sprintf(buffer, "[BP][SI]%x", offset); break;
		case 3:	sprintf(buffer, "[BP][DI]%x", offset); break;
		case 4: sprintf(buffer, "[SI]%x", offset); break;
		case 5: sprintf(buffer, "[DI]%x", offset); break;
		case 6: sprintf(buffer, "[BP]%x", offset); break;
		case 7: sprintf(buffer, "[BX]%x", offset); break;
		default:
			fatal_error();
	}

	return buffer;
}	

static void do_nop(void)
{
	set_string("NOP");	
}
	
static void two_operand_instruction(const char *name, char *(*func1)(), char *(*func2)())
{
	char *tmp;

	process_secondary_byte();
	set_string(name);
	set_string("\t");
	tmp = (*func1)();
	set_string(tmp);

	set_string(",");
	tmp = (*func2)();
	set_string(tmp);
}
	
static char *show_effective_address(BOOLEAN word)
{
	WORD disp;
	static char buffer[20];

	if(mod == 3)
		return word ? show_word_reg(rm) : show_byte_reg(rm);

	if(mod == 0 && rm == 6) {
		disp = read_next_pc_word();
		sprintf(buffer, "%s[%x]", show_seg_override(), disp);
		return buffer;
	}

	sprintf(buffer, "%s%s", show_seg_override(), show_offset());
	return buffer;
}

static char *show_effective_address_word()
{
	return show_effective_address(TRUE);
}

static char *show_effective_address_byte()
{
	return show_effective_address(FALSE);
}

static char *show_seg_reg(unsigned index)
{
	static char *seg_reg_table[] = { "ES", "CS", "SS", "DS" };

	if(index > sizeof(seg_reg_table))
		return "NULL";
	else return seg_reg_table[index];
}

static char *show_seg_reg_from_index(void)
{
	return show_seg_reg(secondary_opcode >> 3); 
}

static char *show_word_reg(int index)
{

	static char *word_reg_table[] = 
		{ "AX", "CX", "DX", "BX", "SP", "BP", "SI", "DI" };

	if(index < 0 || index > sizeof(word_reg_table))
		return "word reg out of range";

	return word_reg_table[index];
}

static char *show_word_reg_from_opcode(void)
{
	return show_word_reg(secondary_opcode >> 3);
}

static char *show_byte_reg(int index)
{
	static char *byte_reg_table[] = 
		{ "AL", "CL", "DL", "BL", "AH", "CH", "DH", "BH" };

	if(index < 0 || index > sizeof(byte_reg_table))
		return "byte reg out of range";

	return byte_reg_table[index];
}

static char *show_byte_reg_from_opcode(void)
{
	return show_byte_reg(secondary_opcode >> 3);
}

static char *show_ea_byte(void)
{
	return show_effective_address(FALSE);	
}

static char *show_ea_word(void)
{
	return show_effective_address(TRUE);
}

static void show_jump(const char *type)
{
	num_bytes++;
	sprintf(disassemble_string, "%s\t%x", type, *pc);
}

static void pop(const char *reg)
{
	sprintf(disassemble_string, "POP\t%s", reg);
}

static void push(const char *reg)
{
	sprintf(disassemble_string, "PUSH\t%s", reg);
}

static void increment(const char *what)
{
	sprintf(disassemble_string, "INC\t%s", what);
}

static void decrement(const char *what)
{
	sprintf(disassemble_string, "DEC\t%s", what);
}

static void reg_to_eabyte(const char *what)
{
	process_secondary_byte();
	sprintf(disassemble_string, "%s\t%s.%s", what, show_ea_byte(), show_byte_reg(secondary_opcode >> 3));

}

static void reg_to_eaword(const char *what)
{
	process_secondary_byte();

	sprintf(disassemble_string, "%s\t%s.%s", what, show_ea_word(), show_word_reg(secondary_opcode >> 3));
}

static void eabyte_to_reg(const char *what)
{
	process_secondary_byte();
	sprintf(disassemble_string, "%s\t%s,%s", what, show_byte_reg(secondary_opcode >> 3), show_ea_byte());

}

static void eaword_to_reg(const char *what)
{
	process_secondary_byte();

	sprintf(disassemble_string, "%s\t%s,%s", what, show_word_reg(secondary_opcode >> 3), show_ea_word());
}

static void imm_to_al_byte(const char *what)
{
}

static void imm_to_ax_word(const char *what)
{
}

static void do_seg_override(const SEG seg)
{
	if(seg_override != NO_SEG) {
		fatal_error();
	}

	seg_override = seg;
	pc++;
	do_disassemble_instruction();
}

static void logical_byte_imm(void)
{
}

static void logical_word_imm(void)
{
}

static void logical_byte_once(void)
{
}

static void logical_word_once(void)
{
}

static void logical_byte_cl(void)
{
}

static void logical_word_cl(void)
{
}

static void do_f6_byte_opcodes(void)
{
}

static void do_f7_word_opcodes(void)
{
}

static void do_fe_byte_opcodes(void)
{
}

static void instruction_with_ea(const char *opcode)
{
	sprintf(disassemble_string, "%s\t%s", show_effective_address_word());
}
 
static void do_ff_word_opcodes(void)
{
	process_secondary_byte();

	switch(secondary_opcode) {
	 	case INC_EA:
			instruction_with_ea("INC");
			break;
		case DEC_EA:
			instruction_with_ea("DEC");
			break;
		case PUSH_EA:
			instruction_with_ea("PUSH");
			break;
		case JMP_NEAR_EA:
			instruction_with_ea("JMP");
			break;
		case JMP_FAR_EA:
			instruction_with_ea("JMPF");
			break;
		case CALL_NEAR_EA:
			instruction_with_ea("CALL");
			break;
		case CALL_FAR_EA:
			instruction_with_ea("CALLF");
			break;
		default:
			sprintf(disassemble_string, "do_ff_word_opcodes == %x\n", secondary_opcode);
			break;
	}
		
}

static char *(*show_func)();

static void set_ea_and_data(const char *opcode,  WORD data)
{
	sprintf(disassemble_string, "%s\t%s,%x", opcode, (*show_func)(),  data);
}

static void do_80_byte_opcodes(void)
{
	process_secondary_byte();

}

static void do_81_word_opcodes(void)
{
	WORD data;

	process_secondary_byte();
	show_func = show_effective_address_word;
	
	data = read_next_pc_word();

	switch(secondary_opcode) {
		case AND:	set_ea_and_data("AND", data); break;
		case OR:	set_ea_and_data("OR", data); break;
		case ADC:	set_ea_and_data("ADC", data); break;
		case ADD:	set_ea_and_data("ADD", data); break;
		case SUB:	set_ea_and_data("SUB", data); break;
		case SBB:	set_ea_and_data("SBB", data); break;
		case XOR:	set_ea_and_data("XOR", data); break;
		case CMP:	set_ea_and_data("CMP", data); break;
		default:	sprintf(disassemble_string, "problem in 81 word opcodes");
	}
}

static void do_83_byte_into_word_opcodes(void)
{
}

static void do_jump_short_imm(void)
{
}

static void do_jump_near_imm(void)
{
	WORD offset = read_next_pc_word();

	sprintf(disassemble_string, "JMP\t%x", offset + current_ip);
}

static void do_jump_far_imm(void)
{
}


static void mov_to_eabyte(void)
{
		
}

static void mov_imm_word_reg(const char *reg)
{
	sprintf(disassemble_string, "MOV\t%s,%s", reg, show_imm_word());
}

static void mov_imm_byte_reg(const char *reg)
{
	sprintf(disassemble_string, "MOV\t%s,%s", reg, show_imm_byte());
}

static void mov_to_eaword()
{
}

static void mov_to_bytereg()
{
	two_operand_instruction("MOV", show_byte_reg_from_opcode, show_ea_byte);
}

static void mov_to_wordreg(void)
{
	two_operand_instruction("MOV", show_word_reg_from_opcode, show_ea_word);
 
}

static void mov_seg_to_ea(void)
{
	two_operand_instruction("MOV", show_seg_reg_from_index, show_ea_word);
}

static void mov_ea_to_seg(void)
{
	two_operand_instruction("MOV", show_ea_word, show_seg_reg_from_index);
}

static void mov_imm_byte_to_ea(void)
{
	two_operand_instruction("MOV", show_imm_byte, show_ea_byte);
}

static void mov_imm_word_to_ea(void)
{
	two_operand_instruction("MOV", show_imm_word, show_ea_word);
}

static void do_movs_byte(void)
{
	set_string("MOVSB");
}

static void do_movs_word(void)
{
	set_string("MOVSW");	
}

static void do_call_near(WORD new_offset)
{
	sprintf(disassemble_string, "CALL\t%x", new_offset + current_ip + 1);	
}


static void do_call_near_imm(void)
{
	do_call_near(read_next_pc_word());
}

static void do_call_far_imm(void)
{
}
	

static void mov_word_ax_to_xb(void) 
{
	WORD address = read_next_pc_word(); 
	sprintf(disassemble_string, "MOV\t[%x],AX", address);

}

static void mov_byte_al_to_xb(void)
{
	WORD address = read_next_pc_word(); 
	sprintf(disassemble_string, "MOV\t[%x],AL", address);
}

static void mov_word_ax_from_xb(void)
{
	WORD address = read_next_pc_word(); 
	sprintf(disassemble_string, "MOV\tAX,[%x]", address);

}

static void mov_byte_al_from_xb(void)
{
	WORD address = read_next_pc_word(); 
	sprintf(disassemble_string, "MOV\tAL,[%x]", address);
}


static void do_disassemble_instruction(void)
{

	num_bytes++;
	switch(*pc++) {
		case POP_AX:	pop("AX"); 	break;
		case POP_BX:	pop("BX");	break;
		case POP_CX:	pop("CX");	break;
		case POP_DX:	pop("DX");	break;
		case POP_DI:	pop("DI");	break;
		case POP_SI:	pop("SI");	break;
		case POP_BP:	pop("BP");	break;
		case POP_SP:	pop("SP");	break;
		case POP_DS:	pop("DS");	break;
		case POP_ES:	pop("ES");	break;
		case POP_SS:	pop("SS");	break;
		case PUSH_AX:	push("AX"); 	break;
		case PUSH_BX:	push("BX");	break;
		case PUSH_CX:	push("CX");	break;
		case PUSH_DX:	push("DX");	break;
		case PUSH_DI:	push("DI");	break;
		case PUSH_SI:	push("SI");	break;
		case PUSH_BP:	push("BP");	break;
		case PUSH_SP:	push("SP");	break;
		case PUSH_DS:	push("DS");	break;
		case PUSH_ES:	push("ES");	break;
		case PUSH_SS:	push("SS");	break;
		case NOP:	do_nop();	break;
		case POPF:	sprintf( disassemble_string, "POPF"); break;
            case PUSHF:
 				sprintf(disassemble_string, "PUSHF"); break;
            case SAHF:
                    break;
            case LAHF:
                    break;
            case INC_AX:
                    increment("AX");
                    break;
            case INC_BX:
                    increment("BX");
                    break;
            case INC_CX:
                    increment("CX");
                    break;
            case INC_DX:
                    increment("DX");
                    break;
            case INC_SI:
                    increment("SI");
                    break;
            case INC_DI:
                    increment("DI");
                    break;
            case INC_BP:
                    increment("BP");
                    break;
            case INC_SP:
                    increment("SP");
                    break;
 
            /* decrement register */
            case DEC_AX:
                    decrement("AX");
                    break;
            case DEC_BX:
                    decrement("BX");
                    break;
            case DEC_CX:
                    decrement("CX");
		    break; 
            case DEC_DX:
                    decrement("DX");
                    break;
            case DEC_SI:
                    decrement("SI");
                    break;
            case DEC_DI:
                    decrement("DI");
                    break;
            case DEC_BP:
                    decrement("BP");
                    break;
            case DEC_SP:
                    decrement("SP");
                    break;
	    case STD:	set_string("STD"); break;
	    case STC:	set_string( "STC"); break;
	  case STI:	set_string( "STI"); break;
	  case CLI:	set_string("CLI");  break;
	  case CLC:	set_string("CLC"); break;
	  case RETURN_FAR: set_string("RETF"); break;
	  case RETURN_NEAR: set_string("RET"); break;
	   
            case SEG_OVERRIDE_CS: do_seg_override(CS_SEG); break;
            case SEG_OVERRIDE_DS: do_seg_override(DS_SEG); break;
            case SEG_OVERRIDE_ES: do_seg_override(ES_SEG); break;
            case SEG_OVERRIDE_SS: do_seg_override(SS_SEG); break;

            case OR_TO_EABYTE:  reg_to_eabyte("OR"); break;
            case OR_TO_EAWORD:  reg_to_eaword("OR"); break;
            case OR_TO_BYTEREG: eabyte_to_reg("OR"); break;
            case OR_TO_WORDREG: eaword_to_reg("OR"); break;
            case OR_TO_AL:      imm_to_al_byte("OR");            break;
            case OR_TO_AX:      imm_to_ax_word("OR");            break;

            case AND_TO_EABYTE:  reg_to_eabyte("AND"); break;
            case AND_TO_EAWORD:  reg_to_eaword("AND"); break;
            case AND_TO_BYTEREG: eabyte_to_reg("AND"); break;
            case AND_TO_WORDREG: eaword_to_reg("AND"); break;
            case AND_TO_AL:      imm_to_al_byte("AND");            break;
            case AND_TO_AX:      imm_to_ax_word("AND");            break;

            case SUB_TO_EABYTE:  reg_to_eabyte("SUB"); break;
            case SUB_TO_EAWORD:  reg_to_eaword("SUB"); break;
            case SUB_TO_BYTEREG: eabyte_to_reg("SUB"); break;
            case SUB_TO_WORDREG: eaword_to_reg("SUB"); break;
            case SUB_TO_AL:      imm_to_al_byte("SUB");            break;
            case SUB_TO_AX:      imm_to_ax_word("SUB");            break;

	    case JO:	show_jump("JO"); break;
	    case JNO:  	show_jump("JNO"); break;
	    case JL:	show_jump("JL"); break;
	    case JNL:	show_jump("JNL"); break;
	    case JLE: 	show_jump("JLE"); break;
	    case JNLE:  show_jump("JNLE"); break;
	    case JC:	show_jump("JC"); break;
	    case JNC:	show_jump("JNC"); break;
	    case JZ:	show_jump("JZ"); break;
	    case JNZ:	show_jump("JNZ"); break;

	    case CALL_NEAR_IMM:	do_call_near_imm(); break;
	    case CALL_FAR_IMM:  do_call_far_imm();  break;
	 
	    case MOV_XB_AX:  mov_word_ax_to_xb(); break;
            case MOV_XB_AL:  mov_byte_al_to_xb(); break;
            case MOV_AX_XB:  mov_word_ax_from_xb(); break;
	    case MOV_AL_XB:  mov_byte_al_from_xb(); break;

            /* ROL/ROR/RCL/RCR/SHLSAL/SHR/SAR opcodes */
            case 0xc0: logical_byte_imm(); break;
            case 0xc1: logical_word_imm(); break;
            case 0xd0: logical_byte_once();   break;
            case 0xd1: logical_word_once();   break;
            case 0xd2: logical_byte_cl();     break;
            case 0xd3: logical_word_cl();     break;
 
            case 0xf6: do_f6_byte_opcodes();  break;
            case 0xf7: do_f7_word_opcodes();   break; 
            case 0xfe: do_fe_byte_opcodes();  break;
            case 0xff: do_ff_word_opcodes();  break;
            case 0x80: do_80_byte_opcodes(); break;
            case 0x81: do_81_word_opcodes(); break;
            case 0x83: do_83_byte_into_word_opcodes(); break;
            
            /* jumps */
            case JUMP_SHORT_IMM: do_jump_short_imm(); break;
            case JUMP_NEAR_IMM:  do_jump_near_imm(); break;
            case JUMP_FAR_IMM:   do_jump_far_imm(); break;

            case MOV_TO_EABYTE: mov_to_eabyte(); break;
            case MOV_TO_EAWORD: mov_to_eaword(); break;
            case MOV_TO_BYTEREG: mov_to_bytereg(); break;
            case MOV_TO_WORDREG: mov_to_wordreg(); break;
            case MOV_SEG_TO_EA:   mov_seg_to_ea(); break;
            case MOV_EA_TO_SEG:   mov_ea_to_seg(); break;
            case MOV_IMM_BYTE_TO_EA: mov_imm_byte_to_ea(); break;
            case MOV_IMM_WORD_TO_EA: mov_imm_word_to_ea(); break;
            case MOVS_BYTE:     do_movs_byte();  break;
            case MOVS_WORD:     do_movs_word(); break;

	    case MOV_AX:	mov_imm_word_reg("AX"); break;
	    case MOV_BX:	mov_imm_word_reg("BX"); break;
	    case MOV_CX:	mov_imm_word_reg("CX"); break;
	    case MOV_DX:	mov_imm_word_reg("DX"); break;
	    case MOV_SI:	mov_imm_word_reg("SI"); break;
	    case MOV_DI:	mov_imm_word_reg("DI"); break;
	    case MOV_BP:	mov_imm_word_reg("BP"); break;
	    case MOV_SP:	mov_imm_word_reg("SP"); break;
	   
	    case MOV_AH:	mov_imm_byte_reg("AH"); break;
	    case MOV_AL:	mov_imm_byte_reg("AL"); break; 
	    case MOV_BH:	mov_imm_byte_reg("BH"); break;
	    case MOV_BL:	mov_imm_byte_reg("BL"); break; 
	    case MOV_CH:	mov_imm_byte_reg("CH"); break;
	    case MOV_CL:	mov_imm_byte_reg("CL"); break; 
	    case MOV_DH:	mov_imm_byte_reg("DH"); break;
	    case MOV_DL:	mov_imm_byte_reg("DL"); break; 
		default:	sprintf(disassemble_string, "undefined opcode");
	}
}

/* dissamble the instruction pointed at by p.
 * Return number of bytes in instruction, 0 for error 
 * Leave dissable instruction in dissamble[].
 *
 */
int disassemble_instruction(BYTE *p, WORD ip)
{
	num_bytes= 0;
	disassemble_string[0] = '\0';
	current_ip = ip;
	seg_override = NO_SEG;
	pc = p;
	do_disassemble_instruction();	
	return num_bytes;
}

int disassemble_range(WORD segment, WORD starting_offset, WORD ending_offset)
{
	void *p;
	int num_bytes;

	while(starting_offset < ending_offset) {
		p = &pc_ram[(segment << 4  ) + starting_offset];
		num_bytes = disassemble_instruction(p, starting_offset);
		printf("%x:%x %s\n", segment, starting_offset, disassemble_string);
		starting_offset += num_bytes;	
	}
}
	
