/* 
 * Dos fcb utilities
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>

#include "misc.h"
#include "8086regs.h"
#include "memory.h"
#include "task.h"
#include "dosprotos.h"

typedef struct {
	char drive;
	char  name[8];
	char extension[3];
	char current_block_number[2];
	char logical_record_size[2];
	char  file_size[4];
	char file_date[2];
	char file_time[2];
} NORMAL_FCB;

typedef struct {
	/* first byte is 0xff */
	unsigned char ff;
	/* reserved */
	char reserved[5];
	char attr_byte;
	NORMAL_FCB normal_fcb;
} EXTENDED_FCB;
	
typedef union {
	NORMAL_FCB normal_fcb;
	EXTENDED_FCB extended_fcb;	
} FCB;

typedef enum { NORMAL, EXTENDED } TYPE_FCB;


static char dos_name_buffer[8 + 3 + 1];	/* fcb style name -- 8 + 3 */

/* what attribute flags are */
#define READ_ONLY 0x1
#define HIDDEN 	  0x2
#define SYSTEM    0x4
#define VOLUME_LABEL 0x8
#define SUBDIRECTORY  0x10
#define ARCHIVE  0x20

static TYPE_FCB type_fcb(const FCB *p)
{
	if(p->extended_fcb.ff == 0xff)
		return EXTENDED;
	else 	return NORMAL;
}

static NORMAL_FCB *get_normal_fcb(const FCB *p)
{
	if(type_fcb(p) == EXTENDED)
		return &(p->extended_fcb.normal_fcb);
	else	return (NORMAL_FCB *) p;
}


static BOOLEAN  bool(BYTE i)
{
	if(i)
		return TRUE;
	else	return FALSE;
}

		
/* 
 *  INT 21 - DOS 1+ - PARSE FILENAME INTO FCB
 *         AH = 29h
 *         AL = parsing options
 *             bit 0: skip leading separators
 *             bit 1: use existing drive number in FCB if no drive is specified,
 *                    instead of setting field to zero
 *             bit 2: use existing filename in FCB if no base name is specified,
 *                    instead of filling field with blanks
 *             bit 3: use existing extension in FCB if no extension is specified,
 *                    instead of filling field with blanks
 *             bits 4-7: reserved (0)
 *         DS:SI -> filename string (both '*' and '?' wildcards OK)
 *         ES:DI -> buffer for unopened FCB
 * Return: AL = result code
 *             00h successful parse, no wildcards encountered
 *             01h successful parse, wildcards present
 *             FFh failed (invalid drive specifier)
 *         DS:SI -> first unparsed character
 *         ES:DI buffer filled with unopened FCB (see AH=0Fh)
 */
void fcb_parse_filename(void)
{
	char filename[20];
	char *cp;
	FCB  *fcb;
	NORMAL_FCB *normal_fcb;
	BOOLEAN skip_leading_seperators;
	BOOLEAN use_existing_drive;
	BOOLEAN use_existing_filename;
	BOOLEAN use_existing_extension;
	/* set if we encounter a wildcard */
	BOOLEAN have_wildcard = FALSE;
	int i;

	strncpy(filename, (char *) (fast_ds + regs.si), sizeof(filename));
	fcb = (FCB *) (fast_es + regs.di);

	if(type_fcb(fcb) == EXTENDED) {
		printf("have extended fcb for parse\n");
		
	}

	normal_fcb = get_normal_fcb(fcb); 

	
	skip_leading_seperators = bool(regs.al & 1);
	use_existing_drive = bool(regs.al & 0x2);
	use_existing_filename = bool(regs.al & 0x3);
	use_existing_extension = bool(regs.al & 0x4);

	/* null terminate if a \r exists */
	for(cp = filename, i = 0; *cp && (*cp > ' ') && i < sizeof(filename) ; i++, cp++)
		;

	*cp = '\0';	/* even if its already null */

	printf("filename parsed = %s\n", filename);

	/* convert to upper case */	
	for(cp = filename; *cp; cp++)
		*cp = toupper(*cp);

	/* check drive */
	if(filename[1] == ':') {
		printf("have drive %c\n", filename[0]);
		cp = filename + 2;
	} else cp = filename;

	if(strcmp(".", cp) == 0) {
		for(i = 0; i < 11; i++)
			normal_fcb->name[i] = ' ';
 
		goto done;
	}
	/* parse filename */
	for(i = 0; i < 8;  cp++) {
		if(*cp == '?') {
			normal_fcb->name[i++] = '?';
			have_wildcard = TRUE;
		} else if(*cp == '.' || *cp == '\0' ) {
			/* add spaces and break */
			while(i < 8)
				normal_fcb->name[i++] = ' ';
			break;
		} else if(*cp == '*') {
			/* add ? and break */
			while(i < 8)
				normal_fcb->name[i++] = '?';
			have_wildcard = TRUE;
			cp++;
			break;
		} else normal_fcb->name[i++] = *cp;
	}

	if(*cp == '.')
		cp++;

	/* fill in extension */
	while(i < 11) {
		if(*cp)
			 normal_fcb->name[i++] = *cp++;
		else normal_fcb->name[i++] = ' ';
	}


done:
	/* FIXME */
	normal_fcb->drive = 0;
	if(have_wildcard == TRUE)
		regs.al = 1;
	else 	regs.al = 0;
#if 0
	/* failrue */
	/* maybe parse extension */
	regs.al = 0xff;
#endif
}

/* move this unix name into dos_name_buffer according to some rules */
static void  make_dos_name(char *unixname)
{
	char *cp;
	char filename[9];
	char ext[4];
	const char *three_spaces = "   ";

	/* get filename */
	if(*unixname == '.') {
		strcpy(filename, "DOT");
		cp = filename + 3;
	} else cp = filename;

	for(;  *unixname && *unixname != '.' && cp - filename < 8; cp++,
			unixname++)
		*cp = toupper(*unixname);
	*cp = '\0';

	if(index(unixname, '.'))
		unixname = index(unixname, '.') + 1;

	for(cp = ext; *unixname && cp - ext < 3; unixname++) {
		if(*unixname == '.')
			continue;
		*cp++ = toupper(*unixname);
	}

	*cp = '\0';
 
	sprintf(dos_name_buffer, "%-8s", filename);
	if(*ext) {
		sprintf(dos_name_buffer + 8, "%-3s", ext);
	}	else	strcat(dos_name_buffer + 8, three_spaces);
}
	

/* Match the name in the fcb with the name in dos_name_buffer.
 * return true if you have a match, false otherwise.
 */
static BOOLEAN match_name(char *fcb_name)
{
	int i;
	char *dos_name = dos_name_buffer;

	if(*fcb_name == '\0') {
		/* skip 8 */
		i = 8;
		fcb_name += 8;
		dos_name += 8;
	} else i = 0;

	for(; i < 11; i++, fcb_name++, dos_name++) {
		if(*fcb_name == '?')
			continue;
		if(*fcb_name != *dos_name++)
			return FALSE;
	}
	return TRUE;
}

/* find a match with the name in  dos_name_buffer -- return NULL for no match */
static struct dirent *find_match(FCB *pfcb)
{
	struct dirent *p;
	char *name;
	NORMAL_FCB *pnormal;
	unsigned char extended_attr_byte = -1; 	/* no extended */	

	if(type_fcb(pfcb) == EXTENDED) {
		/* get attribute byte */
		extended_attr_byte = pfcb->extended_fcb.attr_byte;
		if(extended_attr_byte & VOLUME_LABEL) {
			printf("wanted VOLUME LABEL\n");
			return NULL;
		}
	}	
	
	pnormal = get_normal_fcb(pfcb);

	while(1) {	
		p = readdir(current_task->find_first);
		if(!p)
			return NULL;
		if(!strcmp(p->d_name, "."))
			continue;	/* . */
		if(!strcmp(p->d_name, ".."))
			continue;
		make_dos_name(p->d_name);
		if(match_name(pnormal->name) == TRUE)
			return p;
	}
}



/* copy the name in dos_name_buffer into fcb */
static void copy_name(TYPE_FCB type)
{
	NORMAL_FCB *pfcb;
	char *pname;
	int i;

	if(type == NORMAL)
		pfcb = (NORMAL_FCB *) current_task->dta;
	else	pfcb =  &((FCB *)current_task->dta)->extended_fcb.normal_fcb;

	
	pname = pfcb->name;

	for(i = 0; i < 12; i++)
		*pname++ = dos_name_buffer[i];
}
	

static void fix_dta(struct stat *buf, TYPE_FCB type)
{
	FCB *pfcb;
	NORMAL_FCB *normal_fcb;

	pfcb = (FCB *) current_task->dta;
	if(type == EXTENDED) {
		if(S_ISDIR(buf->st_mode) == TRUE)
			pfcb->extended_fcb.attr_byte = SUBDIRECTORY;
		normal_fcb = &(pfcb->extended_fcb.normal_fcb);
	}	else normal_fcb = pfcb;

	/* fix in file size if not directory */
	if(S_ISDIR(buf->st_mode) == FALSE ) {
		long file_size;
		WORD lower_part;
		WORD upper_part;

		file_size = buf->st_size;
		lower_part = file_size & 0xffff;
		upper_part = file_size >> 16;

		write_pc_word(&(normal_fcb->file_size[0]), lower_part);
		write_pc_word(&(normal_fcb->file_size[2]), upper_part);
	}	
	
	/* fill in date */
	write_pc_word(normal_fcb->file_date, 0x100);
	write_pc_word(normal_fcb->file_time, 0x200);
}

/* current_task->find_first is already opened.
 * get next entry from mask in pfcb into the common dta.
 * If there is any problems, set regs.al to 0xff, if it
 * works set al to 0.
 */
static void get_next_fcb_entry(FCB *pfcb)
{
	struct dirent *p;
	struct stat buf;
	int status;

	if(NULL == current_task->find_first)  {
		DOSLOG("no current_task->find_first");
		regs.al = 0xff;
		return;
	}

	
	p = find_match(pfcb);
	if(NULL == p) {
		regs.al = 0xff;
		return;
	}

	copy_name(type_fcb(pfcb));

	status = stat(p->d_name, &buf);
	assert(status != -1);

	fix_dta(&buf, type_fcb(pfcb));

	regs.al = 0;

}

/*
 *INT 21 - DOS 1+ - FIND FIRST MATCHING FILE USING FCB
 *        AH = 11H
 *        DS:DX -> unopened FCB (see AH=0Fh), may contain '?' wildcards
 *Return: AL = status
 *            00h successful
 *                [DTA] unopened FCB for first matching file
 *            FFh no matching filename, or bad FCB
 *Notes:  the type of the returned FCB depends on whether the input FCB was a
 *          normal or an extended FCB
 *        for extended FCBs with search attribute 08h, the volume label (if any)
 *          will be returned even if the current directory is not the root dir.
 *        DOS 3+ also allows the '*' wildcard
 *        the search FCB must not be modified if AH=12h will be used to continue
 *          searching; DOS 3.3 has set the following parts of the FCB:
 *                 0Ch    BYTE    ???
 *                 0Dh    WORD    directory entry number of matching file
 *                 0Fh    WORD    cluster number of current directory
 *                 11h  4 BYTEs   ???
 *                 15h    BYTE    drive number (1=A:)
 *        at least for DOS 3.3, the unopened FCB in the DTA is actually the drive
 *          number followed by the file's directory entry
 *SeeAlso: AH=12h,AH=1Ah,AH=4Eh,INT 2F/AX=111Bh
 */
void dos_fcb_findfirst(void)
{
	FCB *pfcb;

	pfcb = (FCB *) (fast_ds + regs.dx);

	if(current_task->find_first) { 	
		/* already have find_first? */
		closedir(current_task->find_first);
		DOSLOG("already have opendir");
	}	
	
	current_task->ff_strategy = FF_FCB;
	current_task->find_first = opendir(".");
	get_next_fcb_entry(pfcb);
}

/* 
INT 21 - DOS 1+ - CLOSE FILE USING FCB
        AH = 10h
        DS:DX -> File Control Block (see AH=0Fh)
Return: AL = status
            00h successful
            FFh failed
Notes:  a successful close forces all disk buffers used by the file to be
          written and the directory entry to be updated
        not supported by MS Windows 3.0 DOSX.EXE DOS extender
SeeAlso: AH=0Fh,AH=16h,AH=3Eh

*/
void dos_fcb_close(void)
{
	fprintf(stderr, "Doing fcb close at %x:%x\n", regs.ds, regs.dx);
	regs.al = 0xff;
}

void dos_fcb_findnext(void)
{
	FCB *pfcb;

	pfcb = fast_ds + regs.dx;
	assert(current_task->ff_strategy == FF_FCB);
	get_next_fcb_entry(pfcb);
}

