/* Dos I/O
 *      Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>

#include "misc.h"
#include "8086regs.h"
#include "memory.h"
#include "task.h"
#include "doserrno.h"
#include "dosprotos.h"

static char *rcsid = "$Id: dosio.c,v 1.2 1994/07/22 18:16:34 leisner Exp leisner $";
static char compiled_at[] = COMPILED_AT;

#ifndef NUM_HANDLES
#define NUM_HANDLES 30
#endif

#define DOSIO(x, args... ) { if(dos_show_io == TRUE) fprintf(stderr, x, ##args); }

typedef struct {
	char *path;	/* only for unix systems */
	enum { NO_FS, DOS_FS, UNIX_FS } type;
	char dos_path[64];
} FILE_SYSTEM;



/* number of file systems (normally from A-Z).
 * A, B and C are dos filesystems.
 * D-Z are unix filesystems, where the root (/) is mapped to a Unix path
 */
#define NUM_FILESYSTEMS  26
static FILE_SYSTEM file_systems[NUM_FILESYSTEMS]; 

/* we translate from DOS/psp handles numbers to our own handle
 * indexes in this table 
 */
typedef struct {
	BOOLEAN dos_handle;	/* if TRUE, we have to do all the
				 * work, if FALSE, its a Unix handle/file */
	int	unix_fd;	/* the unix handle for the file */
	char    file_name[80];  /* for deubgging */
	BOOLEAN isatty;	/* so we know whether it is */
	int references;		/* times referenced, 0 should not have owned */
} FILE_HANDLE;

static FILE_HANDLE  handles[NUM_HANDLES];



/* return base of drive for unix, or null */
char *get_drive_path_for_unix(const unsigned drive) 
{
	if(drive <  NUM_FILESYSTEMS) {
		if(file_systems[drive].type == UNIX_FS)
			return file_systems[drive].path;
	}
	return NULL;
}

static void debug_dump_handles(void)
{
	FILE_HANDLE *phandle;
	int i;

	for(i = 0; i < NUM_HANDLES; i++) {
		phandle = &handles[i];
		printf("For handle %d: ", i);
		if(phandle->references) {
			printf("%d references, %s, fd = %d, name = %s\n",
				phandle->references,
				phandle->isatty == TRUE ? "tty" : "file",
				phandle->unix_fd,
				phandle->file_name);
		} else printf("Unusued\n");
	}
}	
	 

void append_unix_name_to_current_drive(const char *name)
{
	FILE_SYSTEM *fs;

	fs = &file_systems[current_default_drive];
	assert(fs->type == UNIX_FS);	
	DOSIO("appending %s to current drive", name);

	if(*name == '/') {
		strcpy(fs->dos_path, name);
	} else  strcat(fs->dos_path, name);
			
}

/* put the current drive into dos format */
char *dos_getcwd_on_specific_drive(const int drive /* 0 is default, 1-n */)
{
	FILE_SYSTEM *p;
	
	assert(drive <= NUM_FILESYSTEMS);

	p = &file_systems[drive ? drive : current_default_drive];
	return p->dos_path; 
	
}
 
/* we have to translate from errno into dos_errno */
static void set_dos_errno(void)
{
	flags.carry = TRUE;

	
}



/* looks through job table for current task and returns first entry which is 0xff
 * (or -1 if none found --- 20 looked at)
 */
static int get_free_dos_handle(void)
{
	char *fd; 
	int i;

	for(i = 0, fd = current_task->jft; i < 20; i++, fd++)
		if(*fd ==  -1)
			return i;

	return -1;
} 

/* looks up the ith entry of jft for current task...
 * Then tries to relate this to handles[...]
 * Returns a handle if one is found, NULL Otherwise 
 */
static FILE_HANDLE *look_up_handle(const int i)
{
	int this_handle;
	FILE_HANDLE *phandle;

	if(i > 20 || i < 0)
		return NULL;
	
	this_handle = *(current_task->jft + i);
	if(this_handle < 0 || this_handle > NUM_HANDLES)
		return NULL;

	phandle = &handles[this_handle];	

	/* check to see if handle opened */
	return phandle;
	
}

/* returns the first free handle or NULL if none found */
static FILE_HANDLE *get_free_handle(void)
{
	FILE_HANDLE *phandle;

	for(phandle = &handles[0]; phandle < &handles[NUM_HANDLES]; phandle++)
		if(phandle->references == 0)
			return phandle;	/* found one */

	return NULL;	/* found none */
}

/* close this handle */
static void close_handle(FILE_HANDLE *phandle) 
{
	phandle->references--;

	if(phandle->dos_handle == FALSE) {
		if(!phandle->references) {
			DOSIO("unix close(%d)\n", phandle->unix_fd);
			close(phandle->unix_fd);
		}
	}
}
	

/* set up the first 3 handles for stdin, stdout and stdaux */
void initialize_handles_for_first_task(const TASK *first_task)
{
	FILE_HANDLE *phandle;
	int i;
	static const char  *const names[] = { "stdin", "stdout", "stderr" };
	static const char *const tty =  "/dev/tty";
	
	/* map first 3 handles to unix handles, with first_task
	 * as owner and handle = number
	 */
	for(phandle = &handles, i = 0; i < 3; i++, phandle++) {
		int fd;

		fd = open(tty, O_RDWR);
		if(fd < 0) {
			fprintf(stderr, "Open on %s failed in %s\n", tty, strerror(errno), 
				__FUNCTION__);
			exit(1);
		}
		phandle->dos_handle = FALSE;
		phandle->unix_fd = fd;
		phandle->isatty = TRUE;
		phandle->references = 1;
		sprintf(phandle->file_name, "%s-%s", names[i], tty);	
		*(first_task->jft + i) = i; 
	}

	/* rest of handles isn't used */
	for( ; phandle < &handles[NUM_HANDLES]; phandle++)
		phandle->references = 0;
	/* process handles 3-20 is 0xff */
	for(i = 3; i < 20; i++) 
		*(first_task->jft + i) = 0xff;
}

/*
 * INT 21 - DOS 2+ - IOCTL - GET DEVICE INFORMATION
 *         AX = 4400h
 *         BX = handle
 * Return: CF clear if successful
 *             DX = device information word
 *                 character device
 *                   14: device driver can process IOCTL requests (see AX=4402h)
 *                   13: output until busy supported
 *                   11: driver supports OPEN/CLOSE calls
 *                    7: set (indicates device)
 *                    6: EOF on input
 *                    5: raw (binary) mode
 *                    4: device is special (uses INT 29)
 *                    3: clock device
 *                    2: NUL device
 *                    1: standard output
 *                    0: standard input
 *                 disk file
 *                   15: file is remote (DOS 3+)
 *                   14: don't set file date/time on closing (DOS 3+)
 *                   11: media not removable
 *                    8: (DOS 4+) generate INT 24 if no disk space on write
 *                    7: clear (indicates file)
 *                    6: file has not been written
 *                  5-0: drive number (0 = A:)
 *             AX destroyed
 *         CF set on error
 *             AX = error code (01h,05h,06h) (see AH=59h)
 * Note:   value in DH corresponds to high byte of device driver's attribute word
 *           if handle refers to a character device
 * SeeAlso: AX=4401h,INT 2F/AX=122Bh
 */
static void get_ioctl(FILE_HANDLE *phandle)
{
	WORD device_status = 0;

	flags.carry = FALSE;

	if(phandle->dos_handle == FALSE) {
		struct stat  statbuf;

		if(fstat(phandle->unix_fd, &statbuf)) {
			flags.carry = TRUE;
			return;
		}

		if(S_ISBLK(statbuf.st_mode) || S_ISCHR(statbuf.st_mode)) {
			device_status |=  (1 << 7);	/* device  */			

			/* we need a better way of knowing isatty? */
			if(phandle->isatty == TRUE)
				device_status |= 3;
		}

		regs.dx = device_status;
	}
}


/*
 * INT 21 - DOS 2+ - IOCTL - SET DEVICE INFORMATION
 *         AX = 4401h
 *         BX = handle (must refer to character device)
 *         DX = device information word (see AX=4400h)
 *             (DH must be zero)
 * Return: CF clear if successful
 *         CF set on error
 *             AX = error code (01h,05h,06h,0Dh) (see AH=59h)
 * SeeAlso: AX=4400h,INT 2F/AX=122Bh
 */

static void set_ioctl(FILE_HANDLE *phandle)
{
}

#define EMULATE_DR_DOS
#ifdef EMULATE_DR_DOS
static void dr_dos_version(void)
{
	regs.ah = 0x10; /* single tasking */
	regs.al = 0x65; /* DR-Dos 5.0 */
	regs.dx = regs.ax;	
}
#endif

/*
--------D-214408-----------------------------
INT 21 - DOS 3.0+ - IOCTL - CHECK IF BLOCK DEVICE REMOVABLE
        AX = 4408h
        BL = drive number (00h = default, 01h = A:, etc)
Return: CF clear if successful
            AX = 0000h if removable
               = 0001h if fixed
        CF set on error
            AX = error code (01h,0Fh) (see AH=59h)
*/
static check_if_block_device_removable(void)
{
	printf("checking removability of device %d\n", regs.bl);
	flags.carry = FALSE;
	regs.ax = 1;
}


void do_ioctl(void)
{
	FILE_HANDLE *phandle;

	flags.carry  = FALSE;
	if(regs.al == 0 || regs.al == 1)	 {
		phandle = look_up_handle(regs.bx);

		if(!phandle) {
			flags.carry = TRUE;
			return;
		}
	}
	
	switch(regs.al) {
		case 0:	get_ioctl(phandle); break;
		case 1: set_ioctl(phandle); break;
		case 8: check_if_block_device_removable();
#ifdef  EMULATE_DR_DOS
		case 0x52: dr_dos_version();
			   break;
		case 0x56: /* dr_dos_hist_buffer(); */
			   break;
		case 0x57: /* dr_dos_share_hiload_control(); */
			   break;
#endif
		default:	fprintf(stderr, "Unknown ioctl %x\n", regs.al);
				flags.carry = TRUE;
				break;
	}
}

/*
 * INT 21 - DOS 2+ - "LSEEK" - SET CURRENT FILE POSITION
 *         AH = 42h
 *         AL = origin of move
 *             00h start of file
 *             01h current file position
 *             02h end of file
 *         BX = file handle
 *         CX:DX = offset from origin of new file position
 * Return: CF clear if successful
 *             DX:AX = new file position in bytes from start of file
 *         CF set on error
 *             AX = error code (01h,06h) (see AH=59h)
 * Notes:  for origins 01h and 02h, the pointer may be positioned before the
 *           start of the file; no error is returned in that case, but subsequent
 *           attempts at I/O will produce errors
 *         if the new position is beyond the current end of file, the file will
 *           be extended by the next write (see AH=40h)
 * SeeAlso: AH=24h,INT 2F/AX=1228h
 */
void do_lseek(void)
{
	FILE_HANDLE *phandle;

	flags.carry = TRUE;
	phandle = look_up_handle(regs.bx);

	if(!phandle) {
		dos_errno = INVALID_HANDLE;
		return;
	}

	if(phandle->dos_handle == FALSE) {
		long offset;
		long return_value;

		offset = (regs.cx << 16) + regs.dx;
		if(regs.al < 2) {
			return_value = lseek(phandle->unix_fd, offset, (int) regs.al);
			if(return_value != -1) {
				regs.dx = return_value >> 16;
				regs.ax = return_value & 0xffff;
				flags.carry = FALSE;			
			}
		} 
	}
}

/* 
 * INT 21 - DOS 2+ - "WRITE" - WRITE TO FILE OR DEVICE
 *         AH = 40h
 *         BX = file handle
 *         CX = number of bytes to write
 *         DS:DX -> data to write
 * Return: CF clear if successful
 *             AX = number of bytes actually written
 *         CF set on error
 *             AX = error code (05h,06h) (see AH=59h)
 * Notes:  if CX is zero, no data is written, and the file is truncated or
 *           extended to the current position
 *         data is written beginning at the current file position, and the file
 *           position is updated after a successful write
 *         the usual cause for AX < CX on return is a full disk
 * BUG:    a write of zero bytes will appear to succeed when it actually failed
 *           if the write is extending the file and there is not enough disk
 *           space for the expanded file (DOS 5.0); one should therefore check
 *           whether the file was in fact extended by seeking to 0 bytes from
 *           the end of the file (INT 21/AX=4202h/CX=0/DX=0)
 *         under the FlashTek X-32 DOS extender, the pointer is in DS:EDX
 * SeeAlso: AH=28h,AH=3Fh,AH=93h,INT 2F/AX=1109h
 * 
 */
void do_write(void)
{

	FILE_HANDLE *phandle =  look_up_handle(regs.bx);

	flags.carry = FALSE;

	DOSIO("write(%d, %s, %d)\n",  regs.bx, fast_ds + regs.dx, regs.cx);

	if(!phandle) {
		flags.carry = TRUE;
		return;
	}
	
	if(phandle->dos_handle == FALSE) {
		regs.ax = write(phandle->unix_fd, fast_ds + regs.dx, regs.cx);
	}

	
	
}

static int low_level_read(FILE_HANDLE *phandle, BYTE *buffer, int sizeof_buffer)
{
	int size;

	if(phandle->dos_handle == FALSE) 
		size = read(phandle->unix_fd, buffer, sizeof_buffer - 1);
	else {
		printf("don't support doshandles\n");
		abort();
	}

	if(size < 0) {
		fprintf(stderr, "bad read from handle %d\n", phandle->unix_fd);
		perror("");
	}
	
	if(phandle->isatty && size) {
		char *cp;

		/* fix \n to \r\n pairs */
		cp = buffer + size -1;
		if(*cp == '\n') {
		    	/* check size ? */
			*cp++ = '\r';
			*cp = '\n';
			size++;
		}
	}
	return size;
}
	
/*
 * ----------213F-------------------------------
 * INT 21 - DOS 2+ - "READ" - READ FROM FILE OR DEVICE
 *         AH = 3Fh
 *         BX = file handle
 *         CX = number of bytes to read
 *         DS:DX -> buffer for data
 * Return: CF clear if successful
 *             AX = number of bytes actually read (0 if at EOF before call)
 *         CF set on error
 *             AX = error code (05h,06h) (see AH=59h)
 * Notes:  data is read beginning at current file position, and the file position
 *           is updated after a successful read
 *         the returned AX may be smaller than the request in CX if a partial
 *           read occurred
 *         if reading from CON, read stops at first CR
 *         under the FlashTek X-32 DOS extender, the pointer is in DS:EDX
 * SeeAlso: AH=27h,AH=40h,AH=93h,INT 2F/AX=1108h,INT 2F/AX=1229h
 * 
 */
void do_read(void)
{

	FILE_HANDLE *phandle =  look_up_handle(regs.bx);

	flags.carry = FALSE;

	if(!phandle) {
		flags.carry = TRUE;
		return;
	}

	regs.ax = low_level_read(phandle, fast_ds + regs.dx, regs.cx);
	
}


/* support both the open and creat call.
 * Assumes carry flag is already SET, , clear it if it works 
 */
static void do_unix_open(const int open_flags, char *unix_name)
{
	int fd;
	FILE_HANDLE *phandle;
	int dos_jft_entry;


	fd = open(unix_name, open_flags, 0666);
	DOSIO("tried to open %s with flags %x, returned %d\n",  unix_name, open_flags, fd);

			
	if(fd < 0) {
			return;
	}
	
	phandle = get_free_handle();
	if(phandle == NULL) {
		dos_messages("No free handles");
		return;
	}
	
	/* put file descriptor into handle */
	phandle->dos_handle = FALSE; 
	phandle->unix_fd = fd;	
	phandle->isatty = isatty(fd) ?  TRUE : FALSE;
	
	dos_jft_entry = get_free_dos_handle();
	if(dos_jft_entry == -1)	{
		/* problem -- clean up */
		return;
	}
	*(current_task->jft + dos_jft_entry) = phandle - handles; 
	regs.ax = dos_jft_entry;
	flags.carry = FALSE;

	phandle->references++;
	
	strcpy(phandle->file_name, unix_name); 	
	dos_messages("unix fd = %d, dos fd = %d\n", fd, dos_jft_entry);
}

/* 
 * INT 21 - DOS 2+ - "OPEN" - OPEN EXISTING FILE
 *         AH = 3Dh
 *         AL = access and sharing modes
 *             bits 2-0: access mode
 *                 000 read only
 *                 001 write only
 *                 010 read/write
 *             bit 3: reserved (0)
 *             bits 6-4: sharing mode (DOS 3+)
 *                 000 compatibility mode
 *                 001 "DENYALL" prohibit both read and write access by others
 *                 010 "DENYWRITE" prohibit write access by others
 *                 011 "DENYREAD" prohibit read access by others
 *                 100 "DENYNONE" allow full access by others
 *             bit 7: inheritance
 *                 if set, file is private to current process and will not be
 *                   inherited by child processes
 *         DS:DX -> ASCIZ filename
 *         CL = attribute mask of files to look for (server call only)
 * Return: CF clear if successful
 *             AX = file handle
 *         CF set on error
 *             AX = error code (01h,02h,03h,04h,05h,0Ch) (see AH=59h)
 * Notes:  file pointer is set to start of file
 *         file handles which are inherited from a parent also inherit sharing
 *           and access restrictions
 *         files may be opened even if given the hidden or system attributes
 *         under the FlashTek X-32 DOS extender, the pointer is in DS:EDX
 * SeeAlso: AH=0Fh,AH=3Ch,AX=4301h,AX=5D00h,INT 2F/AX=1226h
 * 
 * FIXME -- "con" is /dev/tty.
 * 	check for 
 */
void do_open(void)
{
	char *unix_name;
	flags.carry = TRUE;	/* start off assuming error */


	dos_messages("trying to open %s\n", fast_ds + regs.dx);
	unix_name = convert_dos_asciiz_to_unix_name(fast_ds + regs.dx);

	if(unix_name)  {
		int open_flags;

		switch(regs.al & 0x3) {
			case 0:	
				open_flags = O_RDONLY;
				break;
			case 1:
				open_flags = O_WRONLY ;
			 	break;
			case 2:
				open_flags = O_RDWR;
				break;
			default:
				return;
		}
		
		if(open_flags != O_RDONLY)
			open_flags |= O_CREAT;
		do_unix_open(open_flags, unix_name);
		
	} else fprintf(stderr, "Can't open dos files yet\n");			
	
}

/* 
 * INT 21 - DOS 2+ - "CLOSE" - CLOSE FILE
 *         AH = 3Eh
 *         BX = file handle
 * Return: CF clear if successful
 *             AX destroyed
 *         CF set on error
 *             AX = error code (06h) (see AH=59h)
 * Note:   if the file was written to, any pending disk writes are performed, the
 *           time and date stamps are set to the current time, and the directory
 *           entry is updated
 * SeeAlso: AH=10h,AH=3Ch,AH=3Dh,INT 2F/AX=1106h,INT 2F/AX=1227h
 *  
 */
void do_close(void)
{
	FILE_HANDLE *phandle =  look_up_handle(regs.bx);

	flags.carry = FALSE;

	if(!phandle) {
		flags.carry = TRUE;
		return;
	}
	
	DOSIO("close dos handle %d\n", regs.bx);
	phandle->references--;
	if(phandle->dos_handle == FALSE && !phandle->references ) {
		/* close this through unix */
				
		if(close(phandle->unix_fd)) {
			fprintf(stderr, "Can't close %d\n", phandle->unix_fd);
			flags.carry = TRUE;
		}		
	}

	*(current_task->jft + regs.bx) = 0xff;	/* null this out */
		
	
}

/*
 * INT 21 - DOS 2+ - "CREAT" - CREATE OR TRUNCATE FILE
 *         AH = 3CH
 *         CX = file attribute
 *             bit 0: read-only
 *                 1: hidden
 *                 2: system
 *                 3: volume label (ignored)
 *                 4: reserved, must be zero (directory)
 *                 5: archive bit
 *                 7: if set, file is shareable under Novell NetWare
 *         DS:DX -> ASCIZ filename
 * Return: CF clear if successful
 *             AX = file handle
 *         CF set on error
 *             AX = error code (03h,04h,05h) (see AH=59h)
 * Notes:  if a file with the given name exists, it is truncated to zero length
 *         under the FlashTek X-32 DOS extender, the pointer is in DS:EDX
 * SeeAlso: AH=16h,AH=3Dh,AH=5Ah,AH=5Bh,AH=93h
 */
void do_create(void)
{
	
	char *unix_name;

	flags.carry = TRUE;	/* start off assuming error */

	dos_messages("trying to creat %s\n", fast_ds + regs.dx);
	
	unix_name = convert_dos_asciiz_to_unix_name(fast_ds + regs.dx);

	if(unix_name)
		do_unix_open(O_WRONLY | O_CREAT | O_TRUNC, unix_name);
	 else fprintf(stderr, "Can't open dos files yet\n");			
}

/* 
 *
 INT 21 - DOS 2+ - "UNLINK" - DELETE FILE
        AH = 41h
        DS:DX -> ASCIZ filename (no wildcards, but see below)
        CL = attribute mask for deletion (server call only, see below)
Return: CF clear if successful
            AX destroyed (DOS 3.3) AL seems to be drive of deleted file
        CF set on error
            AX = error code (02h,03h,05h) (see AH=59h)
 */
void dos_unlink_asciiz(void)
{
	char *unix_name;	

	flags.carry = TRUE;	/* start off assuming error */

	dos_messages("unlink %s\n", fast_ds + regs.dx);

	unix_name = convert_dos_asciiz_to_unix_name(fast_ds + regs.dx);	
	if(unix_name) {

		if(!unlink(unix_name))
			flags.carry = FALSE;
	}

		
		
}


void boot_filesystems(void)
{
	FILE_SYSTEM *psystem;

	/* A, B and C are DOS */
	for(psystem = &file_systems[0]; psystem <= &file_systems[2];
		psystem++) {
		psystem->path = NULL;	
		psystem->type = DOS_FS;
	}
	/* D: is a unix drive are / */	
	psystem->type = UNIX_FS;
	psystem->path = "/";	/* for now */
	strcpy(psystem->dos_path, "/");
	/* everything else is currently unused */
	for(psystem = &file_systems[4];	/* start with E */
		psystem < &file_systems[NUM_FILESYSTEMS - 1];
		psystem++) {
		psystem->path = NULL;
		psystem->type = NO_FS;	
	}
}

/* close handles owned by this task */
void remove_task_handles(TASK *ptask) 
{
	FILE_HANDLE *phandle;
	char *jft;
	int i;

	TEST_MAGIC(ptask);
	for(jft = ptask->jft, i = 0; i < 20; i++, jft++) {
		if(*jft >= 0) {
			phandle = look_up_handle(*jft);
			assert(phandle);
			close_handle(phandle);
		}
	}
}

/* return the unix style directory of NULL */
char *find_first_directory(char *string)
{
	static char path[128];  /* size ?? */
	char *add = path;
	FILE_SYSTEM *fs;
	char *last_element;
	
	/* look for first letter: */
	if(*(string + 1) == ':') {
		/* first character must be [A-Z] or [a-z] */
		if(isupper(*string))
			fs = &file_systems[*string - 'A'];
		else if(islower(*string))
			fs = &file_systems[*string - 'a'];
		else return NULL;
		
		if(fs->type  == NO_FS)
			return NULL;
		if(fs->type == UNIX_FS) {
			strcpy(path, fs->path);
			add=  path + strlen(path);
		} else return NULL;
		string += 2;
		if(*string == '/' || *string == '\\')
			string++;
		*add++ = '/';
		add = '\0';
	}	
	last_element = rindex(string, '/');
	if(!last_element)
		last_element = rindex(string, '\\');
	if(add == path && !last_element)
		return ".";
	return path;
}



/*
 * INT 21 - DOS 1+ - WRITE STRING TO STANDARD OUTPUT
 *         AH = 09h
 *         DS:DX -> '$'-terminated string
 * Return: AL = 24h (the '$' terminating the string, despite official docs which
 *                 state that nothing is returned) (at least DOS 3.3-5.0)
 * Notes:        ^C/^Break checked, and INT 23 called if pressed
 */
void standard_write_string(void)
{

	FILE_HANDLE *phandle  = look_up_handle(1);
	char buffer[256];
	int size;
	char *destination;
	char *source;
	
	assert(phandle);
	
	for(size = 0, source = fast_ds + regs.dx, destination = buffer;
		destination < buffer + sizeof(buffer) && *source != '$';
		source++, destination++, size++)
		*destination = *source;

	assert(destination < (buffer + sizeof(buffer)));

	if(size == 0)
		return;
	
	/* unix for now */
	write(phandle->unix_fd,  buffer, size - 1);
}

/*
 *INT 21 - DOS 3.0+ - CANONICALIZE FILENAME OR PATH
 *        AH = 60h
 *        DS:SI -> ASCIZ filename or path
 *        ES:DI -> 128-byte buffer for canonicalized name
 *^LReturn: CF set on error
 *            AX = error code
 *                02h invalid component in directory path or drive letter only
 *                03h malformed path or invalid drive letter
 *            ES:DI buffer unchanged
 *        CF clear if successful
 *            AH = 00h
 *            AL = destroyed (00h or 5Ch or last char of current dir on drive)
 *            buffer filled with qualified name of form D:\PATH\FILE.EXT or
 *              \\MACHINE\PATH\FILE.EXT
 *Notes:  the input path need not actually exist
 *        letters are uppercased, forward slashes converted to backslashes,
 *          asterisks converted to appropriate number of question marks, and
 *          file and directory names are truncated to 8.3 if necessary.
 *        '.' and '..' in the path are resolved
 *        filespecs on local drives always start with "d:", those on network
 *          drives always start with "\\"
 *        if path string is on a JOINed drive, the returned name is the one that
 *          would be needed if the drive were not JOINed; similarly for a
 *          SUBSTed, ASSIGNed, or network drive letter.   Because of this, it is
 *          possible to get a qualified name that is not legal under the current
 *          combination of SUBSTs, ASSIGNs, JOINs, and network redirections
 *        under DOS 3.3 through 5.00, a device name is translated differently if
 *          the device name does not have an explicit directory or the directory
 *          is \DEV (relative directory DEV from the root directory works
 *          correctly).  In these cases, the returned string consists of the
 *          unchanged device name and extension appended to the string X:/
 *          (forward slash instead of backward slash as in all other cases) where
 *          X is the default or explicit drive letter.
 *        functions which take pathnames require canonical paths if invoked via
 *          INT 21/AX=5D00h
 *        supported by OS/2 v1.1 compatibility box
 *        NetWare 2.1x does not support characters with the high bit set; early
 *          versions of NetWare 386 support such characters except in this call.
 *          In addition, NetWare returns error code 3 for the path "X:\"; one
 *          should use "X:\." instead.
 *        for DOS 3.3-5.0, the input and output buffers may be the same, as the
 *          canonicalized name is built in an internal buffer and copied to the
 *          specified output buffer as the very last step
 *        for DR DOS 6.0, this function is not automatically called when on a
 *          network.  Device drivers reportedly cannot make this call from their
 *          INIT function.  Using the same pointer for both input and output
 *          buffers is not supported in the April 1992 and earlier versions of
 *          DR DOS
 */
void dos_canonicalize_filename(void)
{
	char *src, *dest;
	
	src = fast_ds + regs.si;
	dest = fast_es + regs.di;

	strcpy(dest, src);
	flags.carry = FALSE;
}


/*
 * INT 21 - DOS 1+ - BUFFERED INPUT
 *         AH = 0Ah
 *         DS:DX -> buffer (see below)
 * Return: buffer filled with user input
 * Notes:  ^C/^Break checked, and INT 23 called if detected
 *         reads from standard input, which may be redirected under DOS 2+
 *         if the maximum buffer size (see below) is set to 00h, this call returns
 *           immediately without reading any input
 * SeeAlso: AH=0Ch,INT 2F/AX=4810h
 * 
 * Format of input buffer:
 * Offset  Size    Description
 *  00h    BYTE    maximum characters buffer can hold
 *  01h    BYTE    (input) number of chars from last input which may be recalled
 *                 (return) number of characters actually read, excluding CR
 *  02h  N BYTEs   actual characters read, including the final carriage return
 */
void dos_buffered_input(void)
{
	BYTE *buffer = fast_ds + regs.dx;
	FILE_HANDLE *phandle = look_up_handle(1);
	int size;
	char *cp;
		
	if(*buffer == 0) {
		*(buffer + 2) = 0;
		return;	
	} 

	assert(phandle);

	/* this may not be right */
	size = read(0, buffer +2, *buffer); 
	if(cp = index(buffer + 2, '\n'))
		*cp = '\r';	
	*(buffer + 1) = size - 1;
}	

/* fix directory to return at most 63 characters, changing / to \ */
char *fix_directory(const char *name)
{
	static char directory[64];
	int i;

	for(i = 0; i < sizeof(directory); i++, name++){
		if(*name == '/')
			directory[i] = '\\';
		else if(islower(*name))
			directory[i] = toupper(*name);
		else  	directory[i] = *name; 
	}
	directory[i] = 0;
	return directory;
}
		


/*
 *INT 21 - DOS 2+ - GET FILE'S DATE AND TIME
 *        AX = 5700h
 *        BX = file handle
 *Return: CF clear if successful
 *            CX = file's time
 *                bits 15-11: hours (0-23)
 *                     10-5:  minutes
 *                      4-0:  seconds/2
 *            DX = file's date
 *                bits 15-9: year - 1980
 *                      8-5: month
 *                      4-0: day
 *        CF set on error
 *            AX = error code (01h,06h) (see AH=59h)
 */
void dos_get_file_date_and_time(void)
{
	FILE_HANDLE *phandle  = look_up_handle(regs.bx);
	if(phandle) {
		if(phandle->dos_handle == FALSE) {
			struct stat buf;
			int status;
			struct tm *the_time;
			extern struct tm *localtime();
	
			status = fstat(phandle->unix_fd, &buf);
			
			the_time = localtime(&buf.st_mtime);
			regs.dx = the_time->tm_mday + (the_time->tm_mon << 4);
			regs.dx += (the_time->tm_year - 10) << 9;
			regs.cx = the_time->tm_sec/2;
			regs.cx += (the_time->tm_min) << 4;
			regs.cx += (the_time->tm_hour) << 11; 
			flags.carry = FALSE;	
		} else assert(0);
	} else flags.carry = TRUE;
	
}

#if 0
/*
INT 21 - DOS 2+ - GET FILE'S DATE AND TIME
        AX = 5700h
        BX = file handle
Return: CF clear if successful
            CX = file's time
                bits 15-11: hours (0-23)
                     10-5:  minutes
                      4-0:  seconds/2
            DX = file's date
                bits 15-9: year - 1980
                      8-5: month
                      4-0: day
        CF set on error
            AX = error code (01h,06h) (see AH=59h)
SeeAlso: AX=5701h
-------------------------------------------------------------------------------
INT 21 - DOS 2+ - SET FILE'S DATE AND TIME
        AX = 5701h
        BX = file handle
        CX = new time (see AX=5700h)
        DX = new date (see AX=5700h)
Return: CF clear if successful
        CF set on error
            AX = error code (01h,06h) (see AH=59h)
SeeAlso: AX=5700h
-------------------------------------------------------------------------------
^LINT 21 - DOS 4.x only - GET ???
        AX = 5702h
        BX = ??? (0000h through 0004h)
        DS:SI -> ???
        ES:DI -> result buffer
        CX = size of result buffer
Return: CX = size of returned data
SeeAlso: AX=5703h,AX=5704h
-------------------------------------------------------------------------------
INT 21 - DOS 4.x only - GET ???
        AX = 5703h
        BX = file handle (only 0000h through 0004h valid)
        DS:SI -> ??? passed through to INT 2F/AX=112Dh
        ES:DI -> result buffer
        CX = size of result buffer
Return: CX = size of returned data
        ES:DI -> zero word (DOS 4.0) if CX >= 2 on entry
SeeAlso: AX=5702h,AX=5704h,INT 2F/AX=112Dh
-------------------------------------------------------------------------------
INT 21 - DOS 4.x only - TRUNCATE OPEN FILE TO ZERO LENGTH
        AX = 5704h
        BX = file handle (only 0000h through 0004h valid)
        DS:SI -> ??? passed through to INT 2F/AX=112Dh
        ES:DI -> result buffer
        CX = size of result buffer
Return: CX = size of returned data
        ES:DI -> zero word (DOS 4.0) if CX >= 2 on entry
SeeAlso: AX=5702h,AX=5703h,INT 2F/AX=112Dh
*/
void dos_get_file_date_and_time(void)
{
	printf("function %s subcode %x\n", __FUNCTION__, regs.al);
}
#endif

/*
 *
INT 21 - DOS 2+ - GET FILE ATTRIBUTES
        AX = 4300h
        DS:DX -> ASCIZ filename
Return: CF clear if successful
            CX = file attributes (see AX=4301h)
            AX = CX (DR-DOS 5.0)
        CF set on error
            AX = error code (01h,02h,03h,05h) (see AH=59h)
Notes:  under the FlashTek X-32 DOS extender, the filename pointer is in DS:EDX
        under DR-DOS 3.41 and 5.0, attempts to change the subdirectory bit are
          simply ignored without an error
SeeAlso: AX=4301h,AX=4310h,AH=B6h,INT 2F/AX=110Fh
 * INT 21 - DOS 2+ - "CHMOD" - SET FILE ATTRIBUTES
 *         AX = 4301h
 *         CX = new file attributes (see below)
 *         DS:DX -> ASCIZ filename
 * Return: CF clear if successful
 *             AX destroyed
 *         CF set on error
 *             AX = error code (01h,02h,03h,05h) (see AH=59h)
 * Notes:  will not change volume label or directory attribute bits, but will
 *           change the other attribute bits of a directory (the directory
 *           bit must be cleared to successfully change the other attributes of a
 *           directory, but the directory will not be changed to a normal file as
 *           a result)
 *         MS-DOS 4.01 reportedly closes the file if it is currently open
 *         for security reasons, the Novell NetWare execute-only bit can never
 *           be cleared; the file must be deleted and recreated
 *         under the FlashTek X-32 DOS extender, the filename pointer is in DS:EDX
 *         DOS 5.0 SHARE will close the file if it is currently open in sharing-
 *           compatibility mode, otherwise a sharing violation critical error is
 *           generated if the file is currently open
 * SeeAlso: AX=4300h,AX=4311h,INT 2F/AX=110Eh
 *  
 * Bitfields for file attributes:
 *  bit 7  shareable (Novell NetWare)
 *  bit 6  unused
 *  bit 5  archive
 *  bit 4  directory
 *  bit 3  volume label
 *         execute-only (Novell NetWare)
 *  bit 2  system
 *  bit 1  hidden
 *  bit 0  read-only
 */
void dos_chmod(void)
{
	char *filename = fast_ds + regs.dx;
	struct stat statbuf;
	int status;

	switch(regs.al) {
		case 0:
			/* make real filename */
			status = stat(filename, &statbuf);
			if(!status) 		 {
				regs.cx = 0;
				if(S_ISDIR(statbuf.st_mode))	
					regs.cx|= (1 << 4);
				if(access(filename, R_OK) && !access(filename, W_OK))
					regs.cx != 1;
				flags.carry = FALSE;
				return;
			}	
			break;
		case 1:
			fprintf(stderr, "set attributes for %s to %x\n", filename,regs.cx);

	}
	flags.carry = TRUE;
}

/*
 * INT 21 - DOS 2+ - "DUP" - DUPLICATE FILE HANDLE
 *         AH = 45h
 *         BX = file handle
 * Return: CF clear if successful
 *             AX = new handle
 *         CF set on error
 *             AX = error code (04h,06h) (see AH=59h)
 * Notes:  moving file pointer for either handle will also move it for the other,
 *           because both will refer to the same system file table
 *         for DOS versions prior to 3.3, file writes may be forced to disk by
 *           duplicating the file handle and closing the duplicate
 */
void dos_dup(void)
{

	FILE_HANDLE *original;

	original = look_up_handle(regs.bx);
	printf("DUP handle %d (unix %d)", regs.bx, original->unix_fd);

	if(!original) {
		fprintf(stderr, "can't look up handle %d\n", regs.bx);
		flags.carry = TRUE;
		return;
	} else {
		int free_slot;
		
		free_slot = get_free_dos_handle();
		if(free_slot < 0) {
			flags.carry = TRUE;
			fprintf(stderr, "can't get free handle in dup\n");
			return;
		}	
		
		printf("dupping %s references %d to %d\n", original->file_name, original->references, free_slot);
		*(current_task->jft + free_slot) = (char) (original - handles);
		original->references++;
		flags.carry = FALSE;
		regs.ax = free_slot;
	} 
}


/* 
 INT 21 - DOS 2+ - "DUP2", "FORCEDUP" - FORCE DUPLICATE FILE HANDLE
        AH = 46h
        BX = file handle
        CX = file handle to become duplicate of first handle
Return: CF clear if successful
        CF set on error
            AX = error code (04h,06h) (see AH=59h)
Notes:  closes file with handle CX if it is still open
        DOS 3.30 hangs if BX=CX on entry
        moving file pointer for either handle will also move it for the other,
          because both will refer to the same system file table
*/
void dos_dup2(void)
{
	FILE_HANDLE *original;
	FILE_HANDLE *dup;

	original = look_up_handle(regs.bx);
	DOSIO("DUP2 handle %d onto %d (unix %d)\n", regs.bx, regs.cx, original->unix_fd);

	if(!original) {
		fprintf(stderr, "can't dup2 handle %d\n", regs.bx);
		flags.carry = TRUE;
		return;
	}

	dup = look_up_handle(regs.cx);
	if(dup)
		close_handle(dup);
	*(current_task->jft + regs.cx) = (char) (original - handles);
	original->references++;
	flags.carry = FALSE;
	
}
	

/* copy the jft table from a parent to a child.
 * Increment references for tty.
 */
void inherit_file_table(TASK *child, TASK *parent)
{
	char *childjft, *parentjft;
	int i;
	
	TEST_MAGIC(child);
	TEST_MAGIC(parent);

	for(childjft = child->jft, parentjft = parent->jft,  i = 0; 
		i < 20;
		i++, childjft++, parentjft++)  {
		*childjft = *parentjft;
		if(*parentjft >= 0) {
			FILE_HANDLE *handle;

			handle = look_up_handle(i);
			assert(handle);
			handle->references++;
		}
	}
		
}


char *convert_dos_asciiz_to_unix_name(char  *asciiz)
{
        static char buffer[256];
        FILE_SYSTEM *fs;
        char *cp;
 
        if(*(asciiz + 1) == ':') {
                if(isascii(*asciiz)) {
                        fs = &file_systems[tolower(*asciiz) - 'a'];
                        asciiz += 2;
                } else return NULL; 
                
                /* fs should be the fs we're on */
                if(fs->type != UNIX_FS) {
                        fprintf(stderr, "Can't convert non-unix names\n");
                        return NULL;
                }       
                strcpy(buffer, fs->path);
                strcat(buffer, "/");
        }       else  {
		if(*asciiz == '/' || *asciiz == '\\') {
			buffer[0] = '/';
                	while(asciiz && (*asciiz == '/' || *asciiz == '\\')) 
                	       	 asciiz++;
		} else buffer[0] = '\0';
        }       
        
        
        cp = rindex(buffer, '/');
        if(cp)
                cp++;
        else cp = buffer;
 
        while(*asciiz) {
                if(*asciiz == '\\')
                        *cp = '/'; 
                else *cp = tolower(*asciiz);
                
                asciiz++;
                cp++;
        }       
        *cp = '\0';
        cp = rindex(buffer, "/");
        if(cp)
                cp++;
        else    cp = buffer;
 
        /*  cp is pointing at file */
        /* check for builtin dos devices */
 
        if(!strcmp(buffer, "con")) {
                strcpy(buffer, "/dev/tty");
        }       
 
        fprintf(stderr, "found file %s\n", buffer);
        return buffer;
 
}
