/*
 *
 * This manages the memory in an ms/dos emulator
 *      Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */


/* 
 * $Log: dosmem.c,v $
 * Revision 1.1  1994/07/22  18:12:55  leisner
 * Initial revision
 *
 */


#include <stdio.h>
#include <errno.h>
#include "misc.h"
#include "8086regs.h"
#include "memory.h"
#include "task.h"
#include "dosprotos.h"

static char *rcsid = "$Id: dosmem.c,v 1.1 1994/07/22 18:12:55 leisner Exp leisner $";

/* unused arenas in ram -- sorted in ascending order */
static MEM_BLOCK *free_list;  

static MEM_BLOCK *dos_list;	/* point to blocks used by dos */

/* number of memory blocks in use */
static int num_mem_blocks = 0;	

/* set when we add a block to the free list (so we know we may have to coalesce */
static BOOLEAN add_free = FALSE;

#ifndef NULL
#define NULL	(void *) 0;
#endif


#define DEBUG_MEM
#ifdef DEBUG_MEM
#include <stdarg.h>

static void coalesce_free_list(void);

static void show_list(FILE *stream, MEM_BLOCK *list)
{
	if(stream) {
		while(list) { 
			fprintf(stream, "segment %4x size %4x\n", 
				list->segment, list->size);
			list = list->next;
		}
	}
}

static void show_memory(void)
{
	fprintf(stderr, "DOS list\n");
	show_list(stderr, dos_list);
	fprintf(stderr, "Free list\n");
	show_list(stderr, free_list);
}
static FILE *mem_debug_stream = NULL;

static const char *tracefile = "memory.trace";
static void make_debug_stream(void)
{
	if(!mem_debug_stream) {
		mem_debug_stream = fopen(tracefile, "w");
		if(mem_debug_stream == NULL) {
			fprintf(stderr, "can't open %s: %s\n", tracefile, strerror(errno));
			exit(1);
		}
	}	

}

/* return a string fro this list */
static char *list_name(const MEM_BLOCK **list)
{
	if(list  == &free_list) 
		return "Free list";
	else if(list == &dos_list)
		return "Alloc list";
	else	return "Unknown list";
}

static char *type_name(TYPE_BLOCK type)
{
	switch(type) {
		case FREE_BLOCK: 	return "free";
		case STACK_BLOCK:	return "stack";
		case ALLOCATED_BLOCK:	return "allocated";
		case UNKNOWN_BLOCK:	return "unknown";
		default:		return "???DEFAULT???";
	}
}



#else
}


static void list_name()
{
}
#endif

static void set_arena_size(MEM_BLOCK *pmem, WORD size)
{
	write_pc_word(pmem->dos_memory + 3, size);
}

static void set_arena_owner(MEM_BLOCK *pmem, WORD owner)
{
	write_pc_word(pmem->dos_memory + 1, owner);
}

static WORD get_arena_owner(MEM_BLOCK *pmem)
{
	return	read_pc_word(pmem->dos_memory + 1);
}

static WORD get_arena_size(MEM_BLOCK *pmem)
{
	return read_pc_word(pmem->dos_memory + 3);
}


/* returns pointer to dos memory this block is used for */
static char *pdos_memory(MEM_BLOCK *p) 
{
	if(p) 
		return &pc_ram[p->segment << 4];
	else	return NULL;
}

void make_arena_header(MEM_BLOCK *pmem, WORD owner)
{
	*(char *) pmem->dos_memory = 0x4d;
	set_arena_size(pmem, pmem->size -1);
	set_arena_owner(pmem, owner);
}

/* set up p->dos_memory to point into pc memory */
static void set_dos_memory(MEM_BLOCK *p)
{	
	p->dos_memory = pdos_memory(p);
}


static void show_dos_memory(FILE *stream)
{
	MEM_BLOCK *p;

	for(p = dos_list; p; p = p->next)
		fprintf(stream, "Block at %x: size = %x clicks  DOS says  Tag %x Owner %x Size %x\n", 
				p->segment, p->size, *(p->dos_memory), get_arena_owner(p),
				get_arena_size(p));
}


static void show_dos_mem_stdout(void)
{
	show_dos_memory(stdout);
}

void show_dos_mem_trace(char *fmt, ...)
{
	va_list args;

	make_debug_stream();

	fprintf(mem_debug_stream, "\n");
	va_start(args, fmt);
	vfprintf(mem_debug_stream, fmt, args);
	va_end(args);
	
	fprintf(mem_debug_stream, "\n");
	show_dos_memory(mem_debug_stream);
	fflush(mem_debug_stream);
}

/* return the memory block if one is found at this segment, or NULL
 * if none found 
 * Adds one to p->segment for search
 */
MEM_BLOCK *find_mem_block(const WORD segment)
{
	MEM_BLOCK *p;

	for(p = dos_list; p &&( p->segment + 1 != segment); p = p->next)
		;

	return p;
}

/* add block p to list in sorted order of click address  -- this is either to the
 * allocated list or free list 
 */
static void add_block(MEM_BLOCK **list, MEM_BLOCK *p)
{
	if(!*list) {
		/* empty list */
		*list = p;
		p->next = NULL;
		p->prev = NULL;
	} else {

		if((*list)->segment > p->segment) {
			/* insert at front of list */
			p->next = *list;
			(*list)->prev = p;
			p->prev = NULL;
			*list = p;
		} else {
			MEM_BLOCK *pblock;		/* insert after pblock (if not 
							 * front of list)   */

			/* insert after pblock */
			for(pblock = *list;  (pblock->next) && (pblock->next)->segment < p->segment;
				pblock = pblock->next)
				;
			p->next = pblock->next;
			p->prev = pblock;
			if(pblock->next)
				(pblock->next)->prev = p;
			pblock->next = p;
		} 


	} 
}

void add_to_dos_list(MEM_BLOCK *p)
{
	add_block(&dos_list, p);
	p->type = ALLOCATED_BLOCK;
}

/* remove block p from list...close up links...don't push block on stack. */
static void remove_block(MEM_BLOCK **list, MEM_BLOCK *p)
{
	if(*list == p)
		*list = p->next;

	if(p->next)
	(p->next)->prev = p->prev;
	if(p->prev)
		(p->prev)->next = p->next;
	 
}
	
static void init_memblocks(void)
{

	free_list = NULL;
	dos_list = NULL;
}

/* blocks are malloced/freed */
/* returns pointer to popped memblock, aborts on error */
static MEM_BLOCK *pop_stack(void)
{
	MEM_BLOCK *p;

	p = malloc(sizeof(MEM_BLOCK));
	if(p == NULL) {
		fprintf(stderr, "free memory stack exhausted\n");
		abort();
	}
	num_mem_blocks++;	
	p->type = UNKNOWN_BLOCK;
	return p;
}	

static void push_stack(MEM_BLOCK *p)
{
	num_mem_blocks--;
	free(p);
}

static MEM_BLOCK *get_block_first_fit(const int clicks)
{
	MEM_BLOCK *p;

	for(p = free_list; p && p->next != NULL && p->size <  clicks; p = p->next)
		;

	if(!p)
		return NULL; 	/* problem? */

	if(p->size < clicks)
		return NULL;	/* problem */

	/* block fits into p -- if clicks < block size, need to pop a block and resize
         * this block -- otherwise remove it from list
         */

	if(p->size == clicks)  {
		remove_block(&free_list, p);
		p->type = UNKNOWN_BLOCK;
		set_dos_memory(p);
		return p;
	} else {
		MEM_BLOCK *new_block;

		new_block = pop_stack();
		new_block->size = clicks;
		new_block->segment = p->segment;  /* address of starting segment */
		p->segment += clicks;	/* advance clicks for the block on free list */
		p->size -= clicks;	/* subtract size taken */
		set_dos_memory(p);
		set_dos_memory(new_block);
		add_free = TRUE;	/* we munged the free list */
		return new_block;
	}	
		 
}

static MEM_BLOCK *get_block_best_fit(const int clicks)
{
	fprintf(stderr, "best fit not yet implemented\n");
	return NULL;
}

static MEM_BLOCK *get_block_last_fit(const int clicks)
{
	fprintf(stderr, "last fit not yet implemented\n");
	return NULL;
}

/* returns  biggest free block on free list,
 * Assmes free list isn't empty.
 */
static MEM_BLOCK *find_biggest_free_block(void)
{
	MEM_BLOCK *pmax, *p;
 
	for(pmax = p = free_list; p;  p = p->next)
		if(p->size > pmax->size)
			pmax = p;	
	
	return pmax;
}

/* get largest block -- remove from free list */
MEM_BLOCK *get_biggest_free_block(void)
{
	MEM_BLOCK *p;
	MEM_BLOCK *pmax;	/* maximum block found so far */	

	if(!free_list)
		return NULL;

	pmax = find_biggest_free_block();
	remove_block(&free_list, pmax);	
	set_dos_memory(pmax);	
	return pmax;
}

/* return a block from the free list of specified clicks (16 bytes/click).
 * return NULL if no block exists 
 */
MEM_BLOCK *get_block(int clicks)
{
	MEM_BLOCK *p;

	coalesce_free_list();
	clicks++;
	switch(mem_alloc_strategy) {
		case FIRST_FIT:	p = get_block_first_fit(clicks);	break;		
		case BEST_FIT:  p = get_block_best_fit(clicks); 	break;
		case LAST_FIT:	p = get_block_last_fit(clicks);	break;
		default:	
			fprintf(stderr, "strageness getting block");
			abort();
	}

	if(p) {
		p->owner = current_task;
		if(current_task) 
			make_arena_header(p, current_task->psp_click);
		else	make_arena_header(p, 0);
	}
	
	return p;
}

/*
 *
 * INT 21 - DOS 2+ - ALLOCATE MEMORY
 *         AH = 48h
 *         BX = number of paragraphs to allocate
 * Return: CF clear if successful
 *             AX = segment of allocated block
 *         CF set on error
 *             AX = error code (07h,08h) (see AH=59h)
 *             BX = size of largest available block
 */
void dos_get_block(void)
{
	MEM_BLOCK *pmem;
	int size;


	size = regs.bx;

	pmem = get_block(size);

	show_dos_mem_trace("dos get block, size = %x, pmem = %x",
			size, pmem);
	if(!pmem) {
		if(!free_list) {
			regs.bx = 0;
		} else {
			MEM_BLOCK  *pmax;

			/* failed -- just report largest free block */
			pmax = find_biggest_free_block();
			assert(pmax);
			regs.bx = pmax->size - 1;
		}
		regs.ax = 0x8;
		flags.carry = TRUE;
	} else {
		/* worked  -- add to dos list */
		add_block(&dos_list, pmem);
		flags.carry = FALSE;
		regs.ax = pmem->segment + 1;
	} 
	
}


static BOOLEAN shrink_block(MEM_BLOCK *p, int size)
{
	MEM_BLOCK *new_block;

	if(p->size < size) {
		fprintf(stderr, "this doesn't make sense, shrinking a bigger block?\n");
		return FALSE;
	} 
	
	add_free = TRUE;

	new_block = pop_stack();

	/* free block + changed block = original block */
	new_block->size = p->size - size;
	p->size = size;
	new_block->segment = p->segment + p->size;
	add_block(&free_list, new_block);
	return TRUE;
}

/* add the specified size from free block to dos_block...if all of free block is
 * used, remove it from the free list
 * size include the arena header
 */
static MEM_BLOCK *grow_block(MEM_BLOCK *dos_block, MEM_BLOCK *free_block, int size)
{
	int need_clicks;

	need_clicks = size - dos_block->size;

	dos_block->size = size;
	if(need_clicks == free_block->size) {
		remove_block(&free_list, free_block);
		push_stack(free_block);
	} else if(need_clicks < free_block->size) {
		free_block->size -= need_clicks;
		free_block->segment += need_clicks;
		set_dos_memory(free_block);
	} else { 
		show_memory();
		fatal_error(); 
	}
}

/* find free block at segment, return NULL if it doesn't exist */
static MEM_BLOCK *find_free_block_at(WORD segment)
{
	MEM_BLOCK *pfree;

	for(pfree = free_list; pfree; pfree = pfree->next)
		if(pfree->segment == segment)
			break;
	fprintf(stderr, "find free block at segment %x is %x\n", segment, pfree);
	return pfree;	
}


/*
 * INT 21 - DOS 2+ - FREE MEMORY
 *         AH = 49h
 *         ES = segment of block to free
 * Return: CF clear if successful
 *         CF set on error
 *             AX = error code (07h,09h) (see AH=59h)
 * Notes:  apparently never returns an error 07h, despite official docs; DOS 3.30
 *           code contains only an error 09h exit
 *         DOS 2.1-5.0 does not coalesce adjacent free blocks when a block is
 *           freed, only when a block is allocated or resized
 * SeeAlso: AH=48h,AH=4Ah
 */
void dos_free_block(void)
{
	fprintf(stderr, "No code in %s, want to free %x \n", __FUNCTION__, regs.es);
	flags.carry =  1;
}

/* 
 * INT 21 - DOS 2+ - RESIZE MEMORY BLOCK
 *         AH = 4Ah
 *         BX = new size in paragraphs
 *         ES = segment of block to resize
 * Return: CF clear if successful
 *         CF set on error
 *             AX = error code (07h,08h,09h) (see AH=59h)
 *             BX = maximum paragraphs available for specified memory block
 * Notes:  under DOS 2.1-5.0, if there is insufficient memory to expand the block
 *           as much as requested, the block will be made as large as possible
 *         DOS 2.1-5.0 coalesces any free blocks immediately following the block
 *           to be resized
 * SeeAlso: AH=48h,AH=49h,AH=83h
 */
void do_resize_block(void)
{
	MEM_BLOCK *pblock;
	WORD target;

	show_dos_mem_trace("Resize block at %x to %x clicks", regs.es, regs.bx);
	/* this task has to own this block... */

	target = regs.es - 1;
		
	for(pblock =  dos_list; pblock; pblock = pblock->next)
		if(pblock->segment == target)
			break;

	if(!pblock) {
		/* block not found */
		flags.carry = TRUE;
		fprintf(stderr, "block not found to resize\n"); 
	} else {
		if(pblock->owner != current_task) {
			flags.carry = TRUE;
			fprintf(stderr, "tried to resize block we don't own\n");
		} else {
			if(pblock->size > regs.bx)  {
				shrink_block(pblock, regs.bx); 	
				flags.carry = FALSE;
			} else {
				MEM_BLOCK *free_block;
				int segment;	/* segment where to start looking */

				/* grow block -- we have to think */
				coalesce_free_list();
				segment = pblock->segment + pblock->size;  /* start ?? */
				free_block = find_free_block_at(segment);
				if(!free_block)  {
					flags.carry = TRUE;				
					regs.bx = pblock->size;
				} else {
					/* can we allocate block? */
					if(pblock->size + free_block->size >= regs.bx) {
					 	grow_block(pblock, free_block, regs.bx + 1);
						flags.carry = FALSE;
					} else	{ /* what is the maximum block? */
						regs.bx = pblock->size + free_block->size;
						flags.carry = TRUE;	
					}
				}
			}
		}
	}	
}

/* remove the blocks owned by this task */
void remove_task_blocks(TASK *ptask)
{
	MEM_BLOCK *pblock;
	MEM_BLOCK *ptmp; 	/* used to make a block go away */

	pblock = dos_list;
	while(pblock) {
		if(pblock->owner == ptask) {
			/* remove this block and add to free list */
			ptmp = pblock;
			pblock = pblock->next;
			remove_block(&dos_list, ptmp);
			add_block(&free_list, ptmp);
			add_free = TRUE;
		} else pblock = pblock->next;
	}
}

/* go through free list and coalesce adjacent blocks */
static void coalesce_free_list(void)
{
	MEM_BLOCK *p = free_list;

	if(add_free == FALSE)
		return; /* don't bother */

	/* go through list while we have one more to possibly coallesce */
	while(p && p->next) {
		if(p->segment + p->size == (p->next)->segment) {
			/* need to maintain this block while it is removed */
			MEM_BLOCK *tmp = p->next;
	
			/* have block to coalesce */
			p->size += (p->next)->size;
			p->next = tmp->next;
			if(tmp->next)
				(tmp->next)->prev = p;
			push_stack(tmp);
		} else {
			p = p->next;			
		}
	}
	add_free = FALSE;
		
}


/* debug routine -- display the environment strings */
static void show_environment(char *environment)
{
	while(*environment) {
		printf("%s\n", environment);
		environment += strlen(environment) + 1;
	}
	
}
	
/* size this environment and move the strings into the new
 * memblock allocated for it
 * A double null signals the end of the environment
 */
MEM_BLOCK *make_environment(char *environment)
{
	int size = 0;
	char *cp;
	int i;
	MEM_BLOCK *p;
	extern char exec_prog_name[];

	show_environment(environment);
	for(cp = environment; *cp; cp = environment + size) {
		size += strlen(cp) + 1;
		fprintf(stderr, "environment string = %s\n", cp);
	}
	

	size += 2;
	/* need a 1 at end of environment ?? */
	size += 2;
	/* round up size to 16 */
	p = get_block((size + 16)/16);

	if(p) {
		/* adjust down for a click */
		for(cp = p->dos_memory + 16; *environment; environment += size  ) {
			strcpy(cp, environment);
			size = strlen(environment) + 1;
			cp += size;
		} 
		*cp++ = '\0';
		/* I don't understand this -- by inspection */
		write_pc_word(cp, 1);
		cp += 2;
#if 0
		/* now this is even weirder -- put the program name into here */
		if(exec_prog_name[0]) {
			char *name;

			name = strlen(&exec_prog_name) + exec_prog_name;
			while(name--) {
				if(name <= exec_prog_name)
					break;
				if(*name == '/' || *name == '\\') {
					name++;
					break;
				}
			}
			strcpy(cp + 2, name);
		}
#else
	strcpy(cp, "h:/foo.bar");
#endif
	}
	show_environment(p->dos_memory);
	return p;	
		
}
	
void boot_memory(void)
{
	MEM_BLOCK *p;

	init_memblocks();
	
	p = pop_stack();
	p->segment = 0x1000;	
	p->size = 0x9000;
	p->next = p->prev = NULL;	/* all alone */
	free_list = p;
	p->type = FREE_BLOCK;	

	mem_alloc_strategy = FIRST_FIT;
}

	

