/*
 *   Manage dos tasks.
 *   This is part of an msdos/ibmpc emulator.
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "misc.h"
#include "regtypes.h"
#include "8086regs.h"
#include "memory.h"
#include "task.h"
#include "dosprotos.h"


/* a running count of the number of tasks in the system */
static int num_task_blocks = 0;	

/* current file being loaded -- FALSE for com, TRUE for exe */
static BOOLEAN exe_file;	

/* when running an exe file, these are these initial values to use
 * for these registers
 */
static struct {
	WORD initial_cs;
	WORD initial_ip;
	WORD initial_ss;
	WORD initial_sp;
} exe_regs;


/* a mapping of PSP to segments... */
typedef struct {
	TASK *task;
	WORD segment;
} PSP_MAP;

static const char PATH[] = "PATH=";

#define MAX_PSPS	20

static PSP_MAP psp_map[MAX_PSPS];

static char *try_to_find_program(char *environment, const char *name);

static void map_psp(TASK *task, WORD segment)
{
	PSP_MAP *p;

	for(p = &psp_map; p < psp_map + MAX_PSPS; p++)
		if(p->task == NULL) {
			p->task = task;
			p->segment = segment;
			task->psp_click = segment;
			return;
		}	
	fprintf(stderr, "Should never get here  in map_psp\n");	
}



static BOOLEAN test_psp(WORD segment)
{
	PSP_MAP *p;

	for(p = &psp_map; p < psp_map + MAX_PSPS; p++)
		if(p->segment == segment)
			return TRUE;

	return FALSE;
}


static void show_a_psp(void *psp)
{
	fprintf(stderr, "INT22 Addr: %x:%x\n", read_pc_word(psp + 0xc), read_pc_word(psp + 0xa));
	fprintf(stderr, "INT23 ADDR: %x: %x\n", read_pc_word(psp + 0x10), read_pc_word(psp + 0xe));
	fprintf(stderr, "Parent PSP: %x\n", read_pc_word(psp + 0x16));
	fprintf(stderr, "environment segment: %x\n", read_pc_word(psp + 0x2c));
}

static void show_all_psps(void)
{
	PSP_MAP *p;

	for(p = &psp_map; p < &psp_map[MAX_PSPS]; p++) {
		TASK *ptask;	

		ptask = p->task;
		if(ptask) {
			printf("PSP %d At segment %x, task = %x\n",
				p - psp_map, p->segment, p->task);
			show_a_psp(ptask->psp);
			printf("Return CS:IP = %x:%x\n", ptask->saved_regs.cs,
						         ptask->saved_regs.ip);
		}
	}
}

/* debug function -- display the psp to stdout */
static void display_psp(WORD segment)
{
	assert(test_psp(segment));
	show_a_psp(get_pcram_seg(segment));
}
	
static void unmap_psp(TASK *task)
{
	PSP_MAP *p;

	TEST_MAGIC(task);
	for(p = &psp_map; p < &psp_map[MAX_PSPS]; p++)
		if(p->task == task) {
			p->task = NULL;
			return;
		}

	fprintf(stderr, "Should never get here in unmap_psp\n");
}

/* actually, blocks are malloced/freed */
static TASK *pop_stack(void)
{
	TASK *p;

	p = malloc(sizeof(TASK));
	if(p == NULL)  {
		fprintf(stderr, "free task stack exhausted\n");
		abort();
	}
	
	num_task_blocks++;
	MAKE_MAGIC(p);	
	return p;

}

static void push_stack(TASK *p)
{
	num_task_blocks--;
	CLEAR_MAGIC(p);
	free(p);
}


/* we are currently running no tasks, put tasks into a stack */
static void init_tasks(void)
{

}	

/* assume command line exists -- check less than 80 characters...
 * slot into dta as [count]line followed by \D.
 */
static void slot_dta(TASK *ptask, const char *command_line)
{
	int length;

	if(!command_line)
		length = 0;
	else length = strlen(command_line);
	if(length > 0x78)
		length = 0x78;
	*ptask->dta = length;		
	strncpy(ptask->dta + 1, command_line, length);
	*(ptask->dta + length + 1) = 0x0;
}

/* set up a new task -- if parent is NULL, get the current path 
 * If parent is 0, this is the first task.
 * The memory block to be used is passed, we have to set up jft, fcb1, fcb2
 * and dta to point into it.
 * If we have a command line, slot it in (count, line, ending byte of
 * 0dh
 */
static void initialize_task(TASK *ptask, TASK *parent, MEM_BLOCK *pmem, const char *command_line )
{
	BYTE *psp;	/* where psp is in pc ram */
	WORD psp_seg;

	
	TEST_MAGIC(ptask);	
	psp_seg = pmem->segment +1;

	psp = get_pcram_seg(psp_seg);
	ptask->psp = psp;
	ptask->jft = psp + 0x18;
	ptask->dta = psp + 0x80;
	ptask->dta_seg = psp_seg;
	ptask->dta_offset = 0x80;
	ptask->ff_strategy = FF_NONE;
	if(!parent) {
		/* first task */
		ptask->jft[0] = 0;
		ptask->jft[1] = 1;
		ptask->jft[2] = 2;
		memset(ptask->jft + 3,  0xff, 18);
				
	} else {
		/* have a parent */
		inherit_file_table(ptask, parent);
	}	
	
	ptask->parent = parent;
	ptask->psp_memory = pmem;

	if(command_line)
		slot_dta(ptask, command_line); 	
	ptask->interrupt_23 = FALSE;
	/* segment of  first byte of unallocated memory */

#if 0
	write_pc_word(ptask->psp + 2, pmem->segment + pmem->size + 1);	
#else
	write_pc_word(ptask->psp + 2, 0xa000);
#endif
	ptask->find_first = NULL;
	map_psp(ptask, psp_seg);
}	

/* pmem points to address of psp 
 * Add 256 bytes to start to load file
 * Return TRUE if it works, FALSE for a problem
 */
static MEM_BLOCK *load_com_file(const int fd, const char *name)
{
	char *pc;	/* where to load into pc ram */
	struct stat statbuf;
	MEM_BLOCK *p;

	
	p = get_biggest_free_block();

	if(p == NULL) {
		fprintf(stderr, "no free blocks\n");
		return NULL;
	}

	pc = p->dos_memory; 
	pc += 256;	/* get past psp */


	/* get size of file to read in */
	if(fstat(fd, &statbuf)) {
		/* can't stat file */
		fprintf(stderr, "can't stat file %s\n", name);
		perror("can't stat com file " );
		return NULL;
	}

	/* size should be rational -- 64*1024 bytes? */
	if(statbuf.st_size > 64*1024) {
		fprintf(stderr, "non-rational com file size = %d\n",
			statbuf.st_size);
		return NULL;
	}

	if(read(fd, pc, statbuf.st_size) != statbuf.st_size) {
		perror("read failed for some reason  loading com file:");
		return NULL;
	}

	return p;
	
	
}

/* malloc space and try to read in exe file...
 * If not, return NULL 
 */
static char *get_exe_file(const int fd, const char *name) 
{
	struct stat statbuf;
	char *buffer;
	int size;
	WORD checksum;
	char *cp;

	if(fstat(fd, &statbuf) == -1) {	
		fprintf(stderr, "can't stat %s\n", name);
		return NULL;
	}

	size = statbuf.st_size;
	if(size > 0x40000) {
		fprintf(stderr, "size is too big at %d for %s\n",
			size, name);
		return NULL;
	}

	buffer = malloc(size);
	if(read(fd, buffer, size) != size) {
		fprintf(stderr, "problem reading %d bytes from %s\n",
			size, name);
		perror("");
		free(buffer);
		return NULL;
	}
#if 0
/* This doesn't seem to be used */
	/* compute checksum */
	for(checksum = 0, cp = buffer; cp <  buffer + size; cp += 2)
		checksum += read_pc_word(cp);

	printf("checksum = %x, 1's complement = %x\n", checksum, ~checksum);	
#endif
	return buffer; 
}

/* copy an executable image into memory (pointed in pc ram by pmem) */
static void copy_image(char *pmem, char *image)
{
	int num_bytes;
	int offset;

	num_bytes  = read_pc_word(image + 4)*512;
	num_bytes -= 512;	/* subtract partial page */
	num_bytes += read_pc_word(image + 2);	/* bytes in last page */
	offset = read_pc_word(image + 8) * 16;	/* header size */

	bcopy(image + offset, pmem, num_bytes - offset);
}

/* load an exe into memory --
 * 
 * Return the memory block where the psp starts, or NULL
 * Do segment fixups...
 * 
 * Format of .EXE file header:
 * Offset  Size    Description
 *  00h  2 BYTEs   .EXE signature, either "MZ" or "ZM" (5A4Dh or 4D5Ah)
 *  02h    WORD    number of bytes in last 512-byte page of executable
 *  04h    WORD    total number of 512-byte pages in executable (includes any
 *                 partial last page)
 *  06h    WORD    number of relocation entries
 *  08h    WORD    header size in paragraphs
 *  0Ah    WORD    minimum paragraphs of memory to allocation in addition to
 *                 executable's size
 *  0Ch    WORD    maximum paragraphs to allocate in addition to executable's size
 *  0Eh    WORD    initial SS relative to start of executable
 *  10h    WORD    initial SP
 *  12h    WORD    checksum (ones complement of sum of all words in executable)
 *  14h    DWORD   initial CS:IP relative to start of executable
 *  18h    WORD    offset within header of relocation table (40h for New EXE)
 *  1Ah    WORD    overlay number (normally 0000h = main program)
 */
static MEM_BLOCK *load_exe_file(int fd, const char *name)
{
	MEM_BLOCK *pmem;
	short signature;
	WORD clicks_offset;	/* place in pc_ram to load the image */
	short num_relocation_entries;
	WORD header_size;		/* in paragraphs or clicks */
	WORD offset_of_relocation_table;
	WORD overlay_number;
	char *image; /* malloced space -- we read it in and
				 * get it over with 
				 */
	int clicks;	/* clicks of block to get */	

	image = get_exe_file(fd, name);
	
	if(image == NULL)  
		return NULL;


	signature = read_pc_word(image);
	if(signature != 0x5a4d && signature != 0x4d5a) {
		fprintf(stderr, "signature on %s isn\'t value = %x\n", name, signature);
		return NULL; 
	}

	num_relocation_entries = read_pc_word(image + 0x6);
	exe_regs.initial_ss = read_pc_word(image + 0xe);
	exe_regs.initial_sp = read_pc_word(image + 0x10);

	exe_regs.initial_ip = read_pc_word(image + 0x14);
	exe_regs.initial_cs = read_pc_word(image + 0x16 );

	offset_of_relocation_table = read_pc_word(image + 0x18);
	overlay_number = read_pc_word(image + 0x1a); 

	/* figure out size of block needed */
	clicks = read_pc_word(image + 4)*32; /* executable size */
	clicks += 16;	/* psp */
	clicks += read_pc_word(image + 0xa);  /* clicks to add -- for stack ? */ 
	pmem = get_block(clicks);
	if(!pmem)
		return pmem;	/* problem */

	clicks_offset = pmem->segment + 1 + 16;	/* for head and psp */
	copy_image(get_pcram_seg(clicks_offset), image);

	/* fix up initial segments */

	exe_regs.initial_cs += clicks_offset; 
	exe_regs.initial_ss += clicks_offset; 

#if 0
	fprintf(stderr, "in %s, %d relocations\n", name, num_relocation_entries);
#endif

	if(num_relocation_entries) {
		/* get relocation table */
	        int  segment, offset;
		char *relocation_table = image + offset_of_relocation_table;
		WORD to_be_fixed;
		BYTE *far_ptr;  /* actually a pointer in pc ram */	

		while(num_relocation_entries--) {
			offset = read_pc_word(relocation_table);
			relocation_table += 2;
			segment = read_pc_word(relocation_table);
			relocation_table += 2;
			segment += clicks_offset;
			far_ptr = get_pcram_seg(segment) + offset;
			to_be_fixed = read_pc_word(far_ptr);
			write_pc_word(far_ptr, clicks_offset + to_be_fixed);
		}	
				
	}
	return pmem;	
}

/* For a 0x4b call, put program in memory, allocate a psp and load the
 * program in.
 * return the psp memory block, or NULL for error
 */
static MEM_BLOCK *load_program(const char *name)
{
	int fd;
	char *cp;
	MEM_BLOCK *p;
	
	cp = rindex(name, '.');
	if(!cp) {
		fprintf(stderr, "No extension loading %s\n", name);
		return FALSE;
	}

	cp++;
	if(!strcmp(cp, "exe"))
		exe_file = TRUE;
	else if(!strcmp(cp, "com"))
		exe_file = FALSE;
	else {
		fprintf(stderr, "Loading %s, neither com nor exe\n", name);
		return NULL;
	}

	fd = open(name, O_RDONLY);
	if(fd == -1) {
		fprintf(stderr, "Can\'t open %s\n", name);
		perror(" ");
		return NULL;
	}

	p = (exe_file == FALSE) ? load_com_file(fd, name) : load_exe_file(fd, name);
	
	close(fd);
	return p;

} 


static char *find_path(const char *environment)
{

	if(!environment)
		return NULL;

	while(*environment) {
		if(strncmp(environment, PATH, strlen(PATH)) == 0)
			return environment + strlen(PATH);
		while(*environment)
			environment++;
		/* environment now points to null */
		environment++;	/* next character may be another null, 
				 * indicating the end of the enviornment
				 */
	}
	return NULL;
}

static char *find_first_program(char *environment, char *name)
{
	char *progname;
	char *extension;
	char tryname[12]; 
	char *path;
#if 0
	const char PATH[] = "PATH";	
	
	/* locate path in enviroment */
	if(environment) {
		while(*environment) {
			if(strncmp(environment, PATH, strlen(PATH)) == 0) 
				break;
			environment += strlen(environment) + 2;	
		}
		if(!*environment)
			path = ".";
		else	path = environment;
	} else path = ".";
#else
	path = find_path(environment);
#endif
	extension = rindex(name, '.');

	if(extension) 
		return try_to_find_program(path, name);

	/* try to paste on .exe and .com */
	if(strlen(name) > 8) 
		return NULL;
	sprintf(tryname, "%s.com", name);
	progname = try_to_find_program(path, tryname);
	if(progname)
		return progname;
	
	/* try another one */
	sprintf(tryname, "%s.exe", name);
	return try_to_find_program(path, tryname);	
			
}

BOOLEAN load_first_task(const char *name, MEM_BLOCK *penvironment, const char *command_line)
{
	int fd;
	BOOLEAN loaded;	
	TASK *ptask;
	MEM_BLOCK *pmem_task;
	char *progname;
	WORD psp_seg;

	if(penvironment)
		progname = find_first_program(penvironment->dos_memory, name);
	else    progname = find_first_program(NULL, name);

	if(!progname) {
		fprintf(stderr, "can't start first program for %s\n", name);
		return FALSE;
	}

	pmem_task = load_program(progname);	

	assert(pmem_task != NULL);

	init_tasks();
	ptask = pop_stack();
	current_task = ptask;
	initialize_task(ptask, NULL, pmem_task, command_line);
	
	/* these have to be owned */
	pmem_task->owner = ptask;
	if(penvironment)
		penvironment->owner = ptask;

        /* add blocks to dos list */
        add_to_dos_list(pmem_task);
	psp_seg = pmem_task->segment + 1;
	make_arena_header(pmem_task, psp_seg);
	if(penvironment) {
	        add_to_dos_list(penvironment);	
		make_arena_header(penvironment, psp_seg);
	}
	if(penvironment) {
		write_pc_word(SEG_ENV(ptask->psp), penvironment->segment + 1);
	}	else	write_pc_word(SEG_ENV(ptask->psp), 0xffff);

	/* load registers */
	switch_task(ptask, 0, 0);	
	
	boot_regs.ds = regs.ds;
	boot_regs.es = regs.es;
	boot_regs.ss = regs.ss;
	boot_regs.cs = regs.cs;
	boot_regs.ip = regs.ip;
	boot_regs.sp = regs.sp;
	show_dos_mem_trace("loaded first task");
}


static void copy_dta(const char *source, char *dest)
{
	/* use strcpy? */
	while(*dest++ = *source++)
		;
}

/* make the current task ptask.
 * This is a file just loaded, so exe is set...
 * ptask->psp_memory points to the memory block for this process 
 * 
 * Pass down the address where support to use for the old task... (if current_task
 * is non-null).
 *
 */
static void switch_task(const TASK *ptask, WORD cs, WORD ip)
{
#if 0
	WORD psp_click = ptask->psp_memory->segment;
#else
	WORD psp_click = ptask->psp_click;
#endif

	TEST_MAGIC(ptask);
	if(current_task) {
		/* set up a termination address */
		void *psp;

		psp = current_task->psp;

		write_pc_word(psp +0xc, cs);
		write_pc_word(psp + 0xa, ip);
	}

	current_task = ptask;
	if(exe_file == FALSE) {
		set_selector(DS_SEG, 	psp_click);	
		set_selector(ES_SEG,	psp_click);
		set_selector(SS_SEG,	psp_click);
		set_selector(CS_SEG,    psp_click);
		regs.ip = 0x100;
		regs.sp = 0xfffe;
	} else {
		set_selector(CS_SEG, exe_regs.initial_cs);
		set_selector(SS_SEG, exe_regs.initial_ss);
		set_selector(DS_SEG, psp_click);
		set_selector(ES_SEG, psp_click);
		regs.ip = exe_regs.initial_ip;
		regs.sp = exe_regs.initial_sp;
	}
}

static void copy_fcb(const char *source, char *dest)
{
	fprintf(stderr, "In copy_fcb\n");
}

#if 0
/* look for name on PATH in enviroment.
 * If we can read it, return pointer to name,
 * otherwise return NULL
 */	
static char *try_to_find_program(char *environment, const char *name)	
#else
/* try to find program given a ;
delimited path and a name (which may have an extension)
 */
static char *try_to_find_program(char *env_path, const char *name)	
#endif
{
	static char path[128];
	char unix_name[128];
	/* set if this is af ull patch name */
	BOOLEAN full_path = FALSE;
	int i;

#if 0
	path[0] = '\0';
	
	/* must exec com or exe file */

	/* look for path in environment */
	while(environment && (strncmp(environment, PATH, strlen(PATH)) != 0)) {
		/* get next one */
		environment += strlen(environment) + 1;
		if(!*environment) 
			break;
	}
#endif

	/* does name start with drive: ? */
	if(*(name + 1) == ':') {
		int drive = tolower(*name) - 'a';
		char *drive_path;

		drive_path = get_drive_path_for_unix(drive);
		assert(drive_path);
		strcpy(path, drive_path);
		name += 2;	
		full_path = TRUE;
	}	

	/* FIXME -- case of /a/b/c/d/e */

	if(*name == '\\' || *name == '/') {
		strcat(path, "/");	 
		name++;
	}

	/* eat slashes */
	while(*name == '\\' || *name == '/') 
		name++;

	/* if no PATH or anyway, try . */
	for(i = 0; i < 14 && *name; i++) {
		unix_name[i] = tolower(*name);
		name++;
	}

	/* null terminate */
	unix_name[i]  = '\0';

	if(access(unix_name, R_OK) == 0)
		return unix_name;	
#if 0
	if(!environment || !*environment)
		return NULL;

	/* have path, get passed PATH= */
	environment += strlen(PATH);


	while(*environment) {
		sprintf(path, "%s/%s",  environment, unix_name);
		if(!access(path, R_OK)) 
			return path;
		environment += strlen(environment) + 1;
	} 
	/* can't find match */

	return NULL;
#else
	if(!env_path)
		return NULL;

	while(*env_path) {
		char *cp = &path[0];

		while(*env_path != ';' && *env_path)  {
			if(*env_path == ':') {
				printf("have a : in path\n");
				return NULL;
			}
			if(*env_path == '\\') {
				printf("have a \\ in path");
				return NULL;
			}
			*cp++ = *env_path++;
		}
		if(env_path == ';')
			env_path++;
		sprintf(path, "%s/%s", path, unix_name);
		if(!access(path, R_OK))
			return path;
		
	}
	
	/* can't find program */
	return NULL;
#endif
}

/* for now */
static void fix_up_name(const char *name, char *string)
{
	strcpy(string, name);
}

/* 
 * INT 21 - DOS 2+ - "EXEC" - LOAD AND/OR EXECUTE PROGRAM
 * 	AH = 4Bh
 * 	AL = type of load
 * 	    00h load and execute
 * 	    01h load but do not execute
 * 	    03h load overlay
 * 	    04h load and execute in background (European MSDOS 4.0 only)
 * 		"Exec & Go" (see also AH=80h)
 * 	DS:DX -> ASCIZ program name (must include extension)
 * 	ES:BX -> parameter block (see below)
 * 	CX = mode (subfunction 04h only)
 * 		0000h child placed in zombie mode after termination
 * 		0001h child's return code discarded on termination
 * Return: CF clear if successful
 * 	    BX,DX destroyed
 * 	    if subfunction 01h, process ID set to new program's PSP; get with
 * 		INT 21/AH=62h
 * 	CF set on error
 * 	    AX = error code (01h,02h,05h,08h,0Ah,0Bh) (see AH=59h)
 * Notes:  DOS 2.x destroys all registers, including SS:SP
 * 	for functions 00h and 01h, the calling process must ensure that there
 * 	  is enough unallocated memory available; if necessary, by releasing
 * 	  memory with AH=49h or AH=4Ah
 * 	for function 01h, the AX value to be passed to the child program is put
 * 	  on top of the child's stack
 * 	for function 03h, DOS assumes that the overlay is being loaded into
 * 	  memory allocated by the caller
 * 	function 01h was undocumented prior to the release of DOS 5.0
 * 	some versions (such as DR-DOS 6.0) check the parameters and parameter
 * 	  block and return an error if an invalid value (such as an offset of
 * 	  FFFFh) is found
 * 	background programs under European MSDOS 4.0 must use the new
 * 	  executable format
 * 	new executables begin running with the following register values
 * 		AX = environment segment
 * 		BX = offset of command tail in environment segment
 * 		CX = size of automatic data segment (0000h = 64K)
 * 		ES,BP = 0000h
 * 		DS = automatic data segment
 * 		SS:SP = initial stack
 * 	  the command tail corresponds to an old executable's PSP:0081h and
 * 	  following, except that the 0Dh is turned into a NUL (00h); new
 * 	  format executables have no PSP
 * 	under the FlashTek X-32 DOS extender, only function 00h is supported
 * 	  and the pointers are passed in DS:EDX and ES:EBX
 * BUGS:   DOS 2.00 assumes that DS points at the current program's PSP
 * 	Load Overlay (subfunction 03h) loads up to 512 bytes too many if the
 * 	  file contains additional data after the actual overlay
 * SeeAlso: AX=4B05h,AH=4Ch,AH=4Dh,AH=64h"OS/2",AH=8Ah,INT 2E
 * 
 */
void dos_exec(void)
{
	char *name;
	/* first word is a segment pointer to the enviroment block
         * (or null means use parents block)
         * next three are far points to the command tail, fcb1, fcb2 
	 */
	BYTE *parameter_block;  
	WORD env_click;
	MEM_BLOCK *environment_block;
	char *dos_environment; /* actual environment in dos memory */
	MEM_BLOCK *psp_block;
	TASK *ptask; 	/* new task */
	BYTE *pdta, *pfcb1, *pfcb2;
	char string[80]; 	

	name = fast_ds + regs.dx; 
	parameter_block = fast_es + regs.bx;

	sprintf(string, "execing %s", name);
	show_dos_mem_trace(string);

	env_click = read_pc_word(parameter_block);
	if(!env_click) {
		/* copy parents enviroment */
		env_click = read_pc_word(SEG_ENV(current_task->psp));		
	} 
		
	dos_environment = &pc_ram[env_click << 4];
	
	switch(regs.al) {
		case 0: 	/* load and execute */
		case 1:		/* only load */
				/* fix up name */
			fix_up_name(name, string);
			if(access(string, R_OK) != 0) {
				/* error */
				flags.carry = TRUE;
				return;
			}
			break;
		default:
			fprintf(stderr, "undefined exec opcode %x\n",  regs.al);
			flags.carry = TRUE;
			return;
	}

	environment_block = make_environment(dos_environment);	
	if(environment_block == NULL) {
		flags.carry = TRUE;
		return;
	}

	psp_block = load_program(name);
	if(psp_block == NULL) {
		fprintf(stderr, "can't allocate psp block for %s\n", name);
		flags.carry = TRUE;
		return;
	}

	ptask = pop_stack();	
	initialize_task(ptask, current_task, psp_block, NULL);
	write_pc_word(SEG_ENV(ptask->psp), environment_block->segment + 1);
	
	/* slot in dta, fcb1, fcb2 */
	pdta = convert_far_pointer(parameter_block + 2);	
	pfcb1 = convert_far_pointer(parameter_block + 6);
	pfcb2 = convert_far_pointer(parameter_block + 10);
	copy_fcb(pfcb1, FCB1(ptask->psp));
	copy_fcb(pfcb2, FCB2(ptask->psp));
	memcpy(ptask->dta, pdta, 0x80);

	flags.carry = FALSE;		/* everything worked so far */
	add_to_dos_list(psp_block);
	add_to_dos_list(environment_block);
	environment_block->owner = psp_block->owner = ptask;
	if(regs.al == 0) {

		/* do a context switch */
		(current_task->saved_regs) = regs;
		(current_task->saved_flags) = flags;
		switch_task(ptask, regs.cs, regs.ip);
	}
	show_dos_mem_trace("end exec");
}

void process_exit(void)
{
	TASK *parent;
	BYTE return_code;
	WORD return_cs;
	WORD return_ip;

	TEST_MAGIC(current_task);
	show_dos_mem_trace("start exit");	
	if(current_task->parent == NULL) {
		fprintf(stderr, "last task\n");
		exit(1); 
	}

	return_code = regs.al;

	unmap_psp(current_task);
	remove_task_blocks(current_task);
	remove_task_handles(current_task);
	if(current_task->find_first)
		closedir(current_task->find_first);
	/* get return address out of psp (so fork works) */
#if 0
	/* use parent? */
	return_cs = read_pc_word(current_task->psp + 0xc);
	return_ip = read_pc_word(current_task->psp + 0xa);
#endif

	parent = current_task->parent;

#if 0 
use parent?
	return_cs = read_pc_word(current_task->psp + 0xc);
	return_ip = read_pc_word(current_task->psp + 0xa);

#endif

	push_stack(current_task);


	current_task = parent;

	return_cs = read_pc_word(current_task->psp + 0xc);
	return_ip = read_pc_word(current_task->psp + 0xa);

	TEST_MAGIC(current_task);
	regs = parent->saved_regs;
	flags = parent->saved_flags;

	/* saved return address */
	regs.cs = return_cs;
	regs.ip = return_ip;

	/* get the return address out of the psp (so my fork works) */
	current_task->return_code = return_code;

	set_selector(DS_SEG, regs.ds);
	set_selector(ES_SEG, regs.es);
	set_selector(SS_SEG, regs.ss);
	set_selector(CS_SEG, regs.cs);

	/* reset interrupt 23 */
	if(current_task->interrupt_23) 
		set_pc_interrupt(0x23, current_task->int23_seg, current_task->int23_off);
	else	set_bad_interrupt(0x23);

	show_dos_mem_trace("end exit");	
	
	
}

void get_process_return_code()
{
	regs.al = current_task->return_code;
	regs.ah = 0;	/* change this */
}

/* 
 * INT 21 - DOS 2+ internal - GET CURRENT PROCESS ID (GET PSP ADDRESS)
 *         AH = 51h
 * Return: BX = segment of PSP for current process
 * Notes:  DOS uses the current PSP address to determine which processes own files
 *           and memory; it corresponds to process identifiers used by other OSs
 *         under DOS 2.x, this function cannot be invoked inside an INT 28h
 *           handler without setting the Critical Error flag
 *         under DOS 3+, this function does not use any of the DOS-internal stacks
 *           and may thus be called at any time, even during another INT 21h call
 *         supported by OS/2 compatibility box
 *         identical to the documented AH=62h
 *         undocumented for DOS 2.x-4.x, but newly documented for 5.0.
 */
void get_current_psp(void)
{
	MEM_BLOCK *psp;

	psp = current_task->psp_memory;
	regs.bx = psp->segment;
}

/* 
 * INT 21 - DOS 2+ internal - SET CURRENT PROCESS ID (SET PSP ADDRESS)
 *         AH = 50h
 *         BX = segment of PSP for new process
 * Notes:  DOS uses the current PSP address to determine which processes own files
 *           and memory; it corresponds to process identifiers used by other OSs
 *         under DOS 2.x, this function cannot be invoked inside an INT 28h
 *           handler without setting the Critical Error flag
 *         under DOS 3+, this function does not use any of the DOS-internal stacks
 *           and may thus be called at any time, even during another INT 21h call
 *         some Microsoft applications use segments of 0000h and FFFFh; although
 *           one should only call this function with valid PSP addresses, any
 *           program hooking it should be prepared to handle invalid addresses
 *         supported by OS/2 compatibility box
 *         this call was undocumented prior to the release of DOS 5.0
 * SeeAlso: AH=26h,AH=51h,AH=62h
 * 
 */
void set_current_psp(void)
{
	PSP_MAP *p;

	for(p = &psp_map; p < &psp_map[MAX_PSPS]; p++)
		if(p->segment  == regs.bx) {
#if 0
			current_task = p->task;	/* what else do we have to do? */
#else
			/* save current registers for return */
			current_task->saved_regs = regs;
			current_task->saved_flags = flags;
			current_task = p->task;
#endif
			TEST_MAGIC(current_task);
			return;
		}	
	fprintf(stderr, "Should never get here  in set_current_psp\n");	
	abort();
}

/*
INT 21 - DOS 2+ internal - CREATE CHILD PSP
        AH = 55h
        DX = segment at which to create new PSP
        SI = (DOS 3+) value to place in memory size field at DX:[0002h]
Return: CF clear if successful
 
Notes:  creates a "child" PSP rather than making an exact copy of the current
          PSP; the new PSP's parent pointer is set to the current PSP and the
          reference count for each inherited file is incremented
        (DOS 2+) sets current PSP to DX
        (DOS 3+) marks "no inherit" file handles as closed in child PSP
SeeAlso: AH=26h,AH=50h
*/
void create_child_psp(void)
{
	TASK *ptask;
	MEM_BLOCK *pmem;

	pmem = find_mem_block(regs.dx);
	if(!pmem) {
		flags.carry = TRUE;
		return;
	}
	ptask = pop_stack();
	initialize_task(ptask, current_task,  pmem, NULL);
	flags.carry = FALSE;
}

